#!/usr/bin/env python
"""
@author Micah Huth
"""

from vpython import canvas, color, arrow, compound, keysdown, rate, norm, \
    sqrt, cos, button, menu, checkbox, slider, wtext, degrees, vector, \
    radians
from roboticstoolbox.backends.VPython.common_functions import \
    get_pose_x_vec, get_pose_y_vec, get_pose_pos, \
    x_axis_vector, y_axis_vector, z_axis_vector
from roboticstoolbox.backends.VPython.grid import GraphicsGrid, create_line, \
    create_segmented_line, create_marker
from enum import Enum
from collections.abc import MutableMapping


class UImode(Enum):  # pragma nocover
    CANVASCONTROL = 1
    TEACHPANEL = 2


class UIMMap(MutableMapping):    # pragma nocover
    def __init__(self, *args, **kwargs):
        self.__dict__.update(*args, **kwargs)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)


class GraphicsCanvas3D:  # pragma nocover
    """
    Set up the scene with initial conditions.
        - White background
        - Width, height
        - Title, caption
        - Axes drawn (if applicable)

    :param height: Height of the canvas on screen (Pixels), defaults to 360.
    :type height: `int`, optional
    :param width: Width of the canvas on screen (Pixels), defaults to 640.
    :type width: `int`, optional
    :param title: Title of the plot. Gets displayed above canvas, defaults
        to ''.
    :type title: `str`, optional
    :param caption: Caption (subtitle) of the plot. Gets displayed below the
        canvas, defaults to ''.
    :type caption: `str`, optional
    :param grid: Whether a grid should be displayed in the plot, defaults
        to `True`.
    :type grid: `bool`, optional
    :param g_col: The RGB grid colour
    :type g_col: `list`, optional
    """

    def __init__(self, height=500, width=888, title='', caption='', grid=True, g_col=None):

        # Create a new independent scene
        self.scene = canvas()

        # Apply the settings
        self.scene.background = color.white
        self.scene.width = width
        self.scene.height = height
        self.scene.autoscale = False

        # Disable default controls
        # Remove shift+mouse panning (not very good controls)
        self.scene.userpan = False
        # Keep zoom controls (scrollwheel)
        self.scene.userzoom = True
        # Keep ctrl+mouse enabled to rotate (keyboard rotation more tedious)
        self.scene.userspin = True

        # Apply HTML title/caption
        if title != '':
            self.scene.title = title

        self.scene.append_to_title(
            '<style>#glowscript h3{margin-bottom: -10px;border-bottom: 2px solid #15578a;padding-bottom: 5px;}#glowscript button{background: #8db9db !important;color: #fff !important;border: 2px solid #5F9ED0;border-radius: 8px;}#glowscript > div:nth-of-type(4n){display:inline-block;margin:20px;vertical-align:top;}#glowscript > div:nth-of-type(4n-3){background:#d9d9d9;padding:20px;border-radius:10px;margin-bottom:10px;}</style>' \
            '<script type="text/javascript">var arrow_keys_handler = function(e) {switch(e.keyCode){ case 37: case 39: case 38:  case 40: case 32: e.preventDefault(); break; default: break;}};window.addEventListener("keydown", arrow_keys_handler, false);</script>' \
            '<script type="text/javascript">$(document).keyup(function(event){if (event.which === 32){event.preventDefault();}});</script>'
            # noqa
        )
        # Disable the arrow keys from scrolling in the browser
        # https://stackoverflow.com/questions/8916620/disable-arrow-key-scrolling-in-users-browser

        # Prevent the space bar from toggling the active checkbox/button/etc
        # (default browser behaviour)
        # https://stackoverflow.com/questions/22280139/prevent-space-button-from-triggering-any-other-button-click-in-jquery

        self.__default_caption = caption
        if caption != '':
            self.scene.caption = caption

        # List of robots currently in the scene
        self.__robots = []
        self.__selected_robot = 0
        # List of joint sliders per robot
        self.__teachpanel = []  # 3D, robot -> joint -> options
        self.__teachpanel_sliders = []
        self.__idx_qlim_min, self.__idx_qlim_max, self.__idx_theta, \
            self.__idx_text = 0, 1, 2, 3
        # Checkbox states
        self.__grid_visibility = grid
        self.__camera_lock = False
        self.__grid_relative = True
        # Screen Shot tally
        self._ss_tally = 0

        # Create the UI
        self.__ui_mode = UImode.CANVASCONTROL
        self.__toggle_button_text_dict = {
            UImode.CANVASCONTROL: "Canvas Controls",
            UImode.TEACHPANEL: "Robot Controls"
        }
        self.__ui_controls = UIMMap()
        self.__add_mode_button()
        self.__setup_ui_controls([])

        # Rotate the camera
        convert_grid_to_z_up(self.scene)
        self.scene.waitfor("draw_complete")  # Ensure camera updates before grid is created

        # Any time a key or mouse is held down, run the callback function
        rate(30)  # 30Hz
        self.scene.bind('keydown', self.__handle_keyboard_inputs)

        # Create the grid, and display if wanted
        self.__graphics_grid = GraphicsGrid(self.scene, colour=g_col)
        if not self.__grid_visibility:
            self.__graphics_grid.set_visibility(False)

    #######################################
    #  Canvas Management
    #######################################
    def clear_scene(self):
        """
        This function will clear the screen of all objects
        """
        # Set all robots variables as invisible
        for robot in self.__robots:
            robot.set_reference_visibility(False)
            robot.set_robot_visibility(False)

        self.scene.waitfor("draw_complete")

        new_list = []
        for name in self.__ui_controls.get('menu_robots').choices:
            new_list.append(name)

        self.__selected_robot = 0
        self.__reload_caption(new_list)

    def grid_visibility(self, is_visible):
        """
        Update the grid visibility in the scene

        :param is_visible: Whether the grid should be visible or not
        :type is_visible: `bool`
        """
        self.__graphics_grid.set_visibility(is_visible)

    def add_robot(self, robot):
        """
        This function is called when a new robot is created. It adds it to
        the drop down menu.

        :param robot: A graphical robot to add to the scene
        :type robot: class:`graphics.graphics_robot.GraphicalRobot`
        """
        # ALTHOUGH THE DOCUMENTATION SAYS THAT MENU CHOICES CAN BE UPDATED,
        # THE PACKAGE DOES NOT ALLOW IT.
        # THUS THIS 'HACK' MUST BE DONE TO REFRESH THE UI WITH AN UPDATED LIST

        # Save the list of robot names
        new_list = []
        for name in self.__ui_controls.get('menu_robots').choices:
            new_list.append(name)
        # Add the new one
        new_list.append(robot.name)

        # Add robot to list
        self.__robots.append(robot)
        self.__selected_robot = len(self.__robots) - 1

        num_options = 4
        #  Add spot for current robot settings
        self.__teachpanel.append([[0] * num_options] * robot.num_joints)

        # Add robot joint sliders
        i = 0
        for joint in robot.joints:
            if joint.qlim[0] == joint.qlim[1]:
                self.__teachpanel[self.__selected_robot][i] = [
                    joint.qlim[0], joint.qlim[1],
                    joint.theta, None]
            else:
                string = "{:.2f} rad ({:.2f} deg)".format(
                    joint.theta, degrees(joint.theta))
                self.__teachpanel[self.__selected_robot][i] = [
                    joint.qlim[0], joint.qlim[1],
                    joint.theta, wtext(text=string)]
            i += 1

        # Refresh the caption
        self.__reload_caption(new_list)

        # Set it as selected
        self.__ui_controls.get('menu_robots').index = \
            len(self.__robots) - 1

        # Place camera based on robots effective radius * 1.25
        if robot.robot is not None:
            radius = sum([abs(link.a) + abs(link.d) for link in robot.robot.links]) * 1.25
            self.scene.camera.pos = vector(radius, radius, radius) + get_pose_pos(robot.joints[1].get_pose())
            self.scene.camera.axis = vector(-radius, -radius, -radius)

    def delete_robot(self, robot):
        """
        This function is called when a new robot is to be deleted
        from the scene.

        :param robot: A graphical robot to add to the scene
        :type robot: class:`graphics.graphics_robot.GraphicalRobot`
        """
        if len(self.__robots) == 0 or robot not in self.__robots:
            return

        robot_index = self.__robots.index(robot)

        # Clear the robot visuals
        self.__robots[robot_index].set_reference_visibility(False)
        self.__robots[robot_index].set_robot_visibility(False)

        # Remove from UI
        new_list = []
        for name in self.__ui_controls.get('menu_robots').choices:
            new_list.append(name)

        del new_list[robot_index]
        del self.__robots[robot_index]
        del self.__teachpanel[robot_index]

        self.__selected_robot = 0
        # Update UI
        self.__reload_caption(new_list)
        # Select the top item
        if len(self.__ui_controls.get('menu_robots').choices) > 0:
            self.__ui_controls.get('menu_robots').index = 0

    def is_robot_in_canvas(self, robot):
        """
        Checks whether the given robot is in the canvas

        :param robot: A graphical robot to add to the scene
        :type robot: class:`graphics.graphics_robot.GraphicalRobot`
        """
        return robot in self.__robots

    def take_screenshot(self, filename):
        """
        Take a screenshot and save it
        """
        self.scene.capture(filename)

    #######################################
    #  UI Management
    #######################################
    def __add_mode_button(self):
        """
        Adds a button to the UI that toggles the UI mode
        """
        btn_text = self.__toggle_button_text_dict.get(
            self.__ui_mode, "Unknown Mode Set")
        btn_text = "<span style='font-size:20px;'>" + btn_text + "</span>"

        btn_toggle = button(bind=self.__toggle_mode, text=btn_text)
        self.__ui_controls.btn_toggle = btn_toggle
        self.scene.append_to_caption('\n')

    def __del_robot(self):
        """
        Remove a robot from the scene and the UI controls
        """
        if len(self.__robots) == 0:
            # Alert the user and return
            self.scene.append_to_caption(
                '<script type="text/javascript">alert'
                '("No robot to delete");</script>')
            return

        # Clear the robot visuals
        self.__robots[self.__selected_robot].set_reference_visibility(False)
        self.__robots[self.__selected_robot].set_robot_visibility(False)

        # Remove from UI
        new_list = []
        for name in self.__ui_controls.get('menu_robots').choices:
            new_list.append(name)

        del new_list[self.__selected_robot]
        del self.__robots[self.__selected_robot]
        del self.__teachpanel[self.__selected_robot]

        self.__selected_robot = 0
        # Update UI
        self.__reload_caption(new_list)
        # Select the top item
        if len(self.__ui_controls.get('menu_robots').choices) > 0:
            self.__ui_controls.get('menu_robots').index = 0

    def __handle_keyboard_inputs(self):
        """
        Pans amount dependent on distance between camera and focus point.
        Closer = smaller pan amount

        A = move left (pan)
        D = move right (pan)
        W = move forward (pan)
        S = move backward (pan)

        <- = rotate left along camera axes (rotate)
        -> = rotate right along camera axes (rotate)
        ^ = rotate up along camera axes (rotate)
        V = rotate down along camera axes (rotate)

        Q = roll left (rotate)
        E = roll right (rotate)

        space = move up (pan)
        shift = move down (pan)

        ctrl + LMB = rotate (Default Vpython)
        """
        # If camera lock, just skip the function
        if self.__camera_lock:
            return

        # Constants
        pan_amount = 0.02  # units
        rot_amount = 1.0  # deg

        # Current settings
        cam_distance = self.scene.camera.axis.mag
        cam_pos = vector(self.scene.camera.pos)
        cam_focus = vector(self.scene.center)

        # Weird manipulation to get correct vector directions.
        # (scene.camera.up always defaults to world up)
        cam_axis = (vector(self.scene.camera.axis))  # X
        cam_side_axis = self.scene.camera.up.cross(cam_axis)  # Y
        cam_up = cam_axis.cross(cam_side_axis)  # Z

        cam_up.mag = cam_axis.mag

        # Get a list of keys
        keys = keysdown()

        # Userspin uses ctrl, so skip this check to avoid changing camera pose
        # while ctrl is held
        if 'ctrl' in keys:
            return

        ######################################################################
        # PANNING
        # Check if the keys are pressed, update vectors as required
        # Changing camera position updates the scene center to
        # follow same changes
        if 'w' in keys:
            cam_pos = cam_pos + cam_axis * pan_amount
        if 's' in keys:
            cam_pos = cam_pos - cam_axis * pan_amount
        if 'a' in keys:
            cam_pos = cam_pos + cam_side_axis * pan_amount
        if 'd' in keys:
            cam_pos = cam_pos - cam_side_axis * pan_amount
        if ' ' in keys:
            cam_pos = cam_pos + cam_up * pan_amount
        if 'shift' in keys:
            cam_pos = cam_pos - cam_up * pan_amount

        # Update camera position before rotation
        # (to keep pan and rotate separate)
        self.scene.camera.pos = cam_pos

        ######################################################################
        # Camera Roll
        # If only one rotation key is pressed
        if 'q' in keys and 'e' not in keys:
            # Rotate camera up
            cam_up = cam_up.rotate(angle=-radians(rot_amount), axis=cam_axis)
            # Set magnitude as it went to inf
            cam_up.mag = cam_axis.mag
            # Set
            self.scene.up = cam_up

        # If only one rotation key is pressed
        if 'e' in keys and 'q' not in keys:
            # Rotate camera up
            cam_up = cam_up.rotate(angle=radians(rot_amount), axis=cam_axis)
            # Set magnitude as it went to inf
            cam_up.mag = cam_axis.mag
            # Set
            self.scene.up = cam_up

        ######################################################################
        # CAMERA ROTATION
        d = cam_distance
        move_dist = sqrt(d ** 2 + d ** 2 - 2 * d * d * cos(
            radians(rot_amount)))  # SAS Cosine

        # If only left not right key
        if 'left' in keys and 'right' not in keys:
            # Calculate distance to translate
            cam_pos = cam_pos + norm(cam_side_axis) * move_dist
            # Calculate new camera axis
            cam_axis = -(cam_pos - cam_focus)
        if 'right' in keys and 'left' not in keys:
            cam_pos = cam_pos - norm(cam_side_axis) * move_dist
            cam_axis = -(cam_pos - cam_focus)
        if 'up' in keys and 'down' not in keys:
            cam_pos = cam_pos + norm(cam_up) * move_dist
            cam_axis = -(cam_pos - cam_focus)
        if 'down' in keys and 'up' not in keys:
            cam_pos = cam_pos - norm(cam_up) * move_dist
            cam_axis = -(cam_pos - cam_focus)

        # Update camera position and axis
        self.scene.camera.pos = cam_pos
        self.scene.camera.axis = cam_axis

    def __reload_caption(self, new_list):
        """
        Reload the UI with the new list of robot names

        :param new_list: The new list to apply to the menu
        :type new_list: `list`
        """
        # Remove all UI elements
        for item in self.__ui_controls:
            if self.__ui_controls.get(item) is None:
                continue
            self.__ui_controls.get(item).delete()
        for item in self.__teachpanel_sliders:
            item.delete()
        self.__teachpanel_sliders = []
        # Restore the caption
        self.scene.caption = self.__default_caption
        # Create the updated caption.
        self.__load_mode_ui(new_list)

    def __load_mode_ui(self, new_list):
        """
        Load the UI menu depending on the current mode

        :param new_list: The new list to apply to the menu
        :type new_list: `list`
        """
        self.__add_mode_button()
        if self.__ui_mode == UImode.CANVASCONTROL:
            self.__setup_ui_controls(new_list)
        elif self.__ui_mode == UImode.TEACHPANEL:
            self.__setup_joint_sliders()
        else:
            self.scene.append_to_caption("UNKNOWN MODE ENTERED\n")

    def __setup_ui_controls(self, list_of_names):
        """
        The initial configuration of the user interface

        :param list_of_names: A list of names of the robots in the screen
        :type list_of_names: `list`
        """

        ##################################################
        self.scene.append_to_caption('<h3>Scene Settings</h3>\n')
        # Button to reset camera
        reset_button = button(
            bind=self.__reset_camera, text="Reset Camera")
        self.__ui_controls.btn_reset = reset_button
        self.scene.append_to_caption('\t')

        screenshot_button = button(bind=self.__screenshot, text="Take Screenshot")
        self.__ui_controls.btn_ss = screenshot_button
        self.scene.append_to_caption('\n')

        camera_lock_checkbox = checkbox(bind=self.__camera_lock_checkbox, text="Camera Lock",
                                        checked=self.__camera_lock)
        self.__ui_controls.chkbox_cam = camera_lock_checkbox
        self.scene.append_to_caption('\t')

        grid_relative_checkbox = checkbox(
            bind=self.__grid_relative_checkbox,
            text="Grid Relative", checked=self.__grid_relative)
        self.__ui_controls.chkbox_rel = grid_relative_checkbox

        self.scene.append_to_caption('\t')
        # Checkbox for grid visibility
        checkbox_grid_visibility = checkbox(
            bind=self.__grid_visibility_checkbox, text="Grid Visibility",
            checked=self.__grid_visibility)
        self.__ui_controls.chkbox_grid = checkbox_grid_visibility
        self.scene.append_to_caption('\n')

        ##################################################
        self.scene.append_to_caption('<h3>Robot</h3>\n')
        # Drop down for robots / joints in frame
        menu_robots_list = menu(bind=self.__menu_item_chosen, choices=list_of_names)
        if not len(list_of_names) == 0:
            menu_robots_list.index = self.__selected_robot
        self.__ui_controls.menu_robots = menu_robots_list
        self.scene.append_to_caption('\t')

        # Button to delete the selected robot
        delete_button = button(bind=self.__del_robot, text="Delete Robot")
        self.__ui_controls.btn_del = delete_button
        self.scene.append_to_caption('\t')

        # Button to clear the robots in screen
        clear_button = button(bind=self.clear_scene, text="Clear Scene")
        self.__ui_controls.btn_clr = clear_button
        self.scene.append_to_caption('\n')

        ##################################################
        self.scene.append_to_caption('<h3>Characteristics</h3>\n')
        # Checkbox for reference frame visibilities
        if len(self.__robots) == 0:
            reference_checkbox = checkbox(
                bind=self.__reference_frame_checkbox,
                text="Show Reference Frames", checked=True)
        else:
            chk = self.__robots[self.__selected_robot].ref_shown
            reference_checkbox = checkbox(
                bind=self.__reference_frame_checkbox,
                text="Show Reference Frames", checked=chk)
        self.__ui_controls.chkbox_ref = reference_checkbox
        self.scene.append_to_caption('\t')

        # Checkbox for robot visibility
        if len(self.__robots) == 0:
            robot_vis_checkbox = checkbox(
                bind=self.__robot_visibility_checkbox,
                text="Show Robot", checked=True)
        else:
            chk = self.__robots[self.__selected_robot].rob_shown
            robot_vis_checkbox = checkbox(
                bind=self.__robot_visibility_checkbox,
                text="Show Robot", checked=chk)
        self.__ui_controls.chkbox_rob = robot_vis_checkbox
        self.scene.append_to_caption('\n')

        # Slider for robot opacity
        self.scene.append_to_caption('Robot Opacity:')
        if len(self.__robots) == 0:
            opacity_slider = slider(bind=self.__opacity_slider, value=1)
        else:
            opc = self.__robots[self.__selected_robot].opacity
            opacity_slider = slider(bind=self.__opacity_slider, value=opc)
        self.__ui_controls.sld_opc = opacity_slider
        self.scene.append_to_caption('\n')

        ##################################################
        # Control manual
        controls_str = '<h3>Controls</h3><br>' \
                       '<b>PAN</b><br>' \
                       'W , S | <i>forward / backward</i><br>' \
                       'A , D | <i>left / right</i><br>' \
                       'SPACE , SHIFT | <i>up / down</i><br>' \
                       '<b>ROTATE</b><br>' \
                       'CTRL + LMB | <i>free spin</i><br>' \
                       'ARROWS KEYS | <i>rotate direction</i><br>' \
                       'Q , E | <i>roll left / right</i><br>' \
                       '<b>ZOOM</b></br>' \
                       'MOUSEWHEEL | <i>zoom in / out</i>'

        self.scene.append_to_caption(controls_str)

    def __setup_joint_sliders(self):
        """
        Display the Teachpanel mode of the UI
        """
        self.scene.append_to_caption('\n')
        if len(self.__teachpanel) == 0:
            self.scene.append_to_caption("No robots available\n")
            return
        i = 0
        for joint in self.__teachpanel[self.__selected_robot]:
            if joint[self.__idx_qlim_min] == joint[self.__idx_qlim_max]:
                # If a slider with (effectively) no values, skip it
                i += 1
                continue
            # Add a title
            self.scene.append_to_caption('Joint {0}:\t'.format(i))
            # Add the slider, with the correct joint variables
            s = slider(
                bind=self.__joint_slider,
                min=joint[self.__idx_qlim_min],
                max=joint[self.__idx_qlim_max],
                value=joint[self.__idx_theta],
                id=i
            )
            self.__teachpanel_sliders.append(s)
            string = "{:.2f} rad ({:.2f} deg)".format(
                joint[self.__idx_theta], degrees(joint[self.__idx_theta]))
            joint[self.__idx_text] = wtext(text=string)
            self.scene.append_to_caption('\n\n')
            i += 1

    #######################################
    # UI CALLBACKS
    #######################################
    def __toggle_mode(self):
        """
        Callback for when the toggle mode button is pressed
        """
        # Update mode
        # Update mode, default canvas controls
        self.__ui_mode = {
            UImode.CANVASCONTROL: UImode.TEACHPANEL,
            UImode.TEACHPANEL: UImode.CANVASCONTROL
        }.get(self.__ui_mode, UImode.CANVASCONTROL)

        # Update UI
        # get list of robots
        new_list = []
        for name in self.__ui_controls.get('menu_robots').choices:
            new_list.append(name)

        self.__reload_caption(new_list)

    def __reset_camera(self):
        """
        Reset the camera to a default position and orientation
        """
        # Reset Camera
        self.scene.up = z_axis_vector
        self.scene.camera.pos = vector(10, 10, 10)
        self.scene.camera.axis = -self.scene.camera.pos

        # Update grid
        self.__graphics_grid.update_grid()

    def __menu_item_chosen(self, m):
        """
        When a menu item is chosen, update the relevant checkboxes/options

        :param m: The menu object that has been used to select an item.
        :type: class:`menu`
        """
        # Get selected item
        self.__selected_robot = m.index

        # Update the checkboxes/sliders for the selected robot
        self.__ui_controls.get('chkbox_ref').checked = \
            self.__robots[self.__selected_robot].ref_shown

        self.__ui_controls.get('chkbox_rob').checked = \
            self.__robots[self.__selected_robot].rob_shown

        self.__ui_controls.get('sld_opc').value = \
            self.__robots[self.__selected_robot].opacity

    def __reference_frame_checkbox(self, c):
        """
        When a checkbox is changed for the reference frame option, update the
        graphics

        :param c: The checkbox that has been toggled
        :type c: class:`checkbox`
        """
        if len(self.__robots) > 0:
            self.__robots[self.__selected_robot].set_reference_visibility(
                c.checked)

    def __robot_visibility_checkbox(self, c):
        """
        When a checkbox is changed for the robot visibility, update the
        graphics

        :param c: The checkbox that has been toggled
        :type c: class:`checkbox`
        """
        if len(self.__robots) > 0:
            self.__robots[self.__selected_robot].set_robot_visibility(
                c.checked)

    def __grid_visibility_checkbox(self, c):
        """
        When a checkbox is changed for the grid visibility, update the graphics

        :param c: The checkbox that has been toggled
        :type c: class:`checkbox`
        """
        self.grid_visibility(c.checked)
        self.__grid_visibility = c.checked

    def __camera_lock_checkbox(self, c):
        """
        When a checkbox is changed for the camera lock, update the camera

        :param c: The checkbox that has been toggled
        :type c: class:`checkbox`
        """
        # Update parameters
        # True = locked
        self.__camera_lock = c.checked
        # True = enabled
        self.scene.userspin = not c.checked
        self.scene.userzoom = not c.checked

    def __grid_relative_checkbox(self, c):
        """
        When a checkbox is changed for the grid lock, update the grid

        :param c: The checkbox that has been toggled
        :type c: class:`checkbox`
        """
        self.__graphics_grid.set_relative(c.checked)
        self.__grid_relative = c.checked

    def __opacity_slider(self, s):
        """
        Update the opacity slider depending on the slider value

        :param s: The slider object that has been modified
        :type s: class:`slider`
        """
        if len(self.__robots) > 0:
            self.__robots[self.__selected_robot].set_transparency(s.value)

    def __joint_slider(self, s):
        """
        The callback for when a joint slider has changed value

        :param s: The slider object that has been modified
        :type s: class:`slider`
        """
        # Save the value
        self.__teachpanel[self.__selected_robot][s.id][self.__idx_theta] = \
            s.value

        # Get all angles for the robot
        angles = []
        for idx in range(len(self.__teachpanel_sliders)):
            angles.append(self.__teachpanel_sliders[idx].value)

        # Run fkine
        poses = self.__robots[self.__selected_robot].fkine(angles)

        # Update joints
        self.__robots[self.__selected_robot].set_joint_poses(poses)

        for joint in self.__teachpanel[self.__selected_robot]:
            if joint[self.__idx_text] is None:
                continue
            string = "{:.2f} rad ({:.2f} deg)".format(
                joint[self.__idx_theta], degrees(joint[self.__idx_theta]))
            joint[self.__idx_text].text = string

    def __screenshot(self, b):
        """
        Take a screencap
        """
        filename = "vp_ss_{:04d}.png".format(self._ss_tally)
        self.take_screenshot(filename)
        self._ss_tally += 1


class GraphicsCanvas2D:  # pragma nocover
    """
    Set up the scene with initial conditions.
        - White background
        - Width, height
        - Title, caption
        - Axes drawn (if applicable)

    :param height: Height of the canvas on screen (Pixels), defaults to 360.
    :type height: `int`, optional
    :param width: Width of the canvas on screen (Pixels), defaults to 640.
    :type width: `int`, optional
    :param title: Title of the plot. Gets displayed above canvas, defaults to ''.
    :type title: `str`, optional
    :param caption: Caption (subtitle) of the plot. Gets displayed below the canvas, defaults to ''.
    :type caption: `str`, optional
    :param grid: Whether a grid should be displayed in the plot, defaults to `True`.
    :type grid: `bool`, optional
    :param g_col: The RGB grid colour
    :type g_col: `list`, optional
    """

    def __init__(self, height=500, width=888, title='', caption='', grid=True, g_col=None):

        # Private lists
        self.__line_styles = [
            '',  # None
            '-',  # Solid (default)
            '--',  # Dashes
            ':',  # Dotted
            '-.',  # Dash-dot
        ]
        self.__marker_styles = [
            '+',  # Plus
            'o',  # Circle
            '*',  # Star
            '.',  # Dot
            'x',  # Cross
            's',  # Square
            'd',  # Diamond
            '^',  # Up triangle
            'v',  # Down triangle
            '<',  # Left triangle
            '>',  # Right triangle
            'p',  # Pentagon
            'h',  # Hexagon
        ]
        self.__colour_styles = [
            'r',  # Red
            'g',  # Green
            'b',  # Blue
            'y',  # Yellow
            'c',  # Cyan
            'm',  # Magenta
            'k',  # Black (default)
            'w',  # White
        ]
        self.__colour_dictionary = {
            'r': color.red.value,
            'g': color.green.value,
            'b': color.blue.value,
            'c': color.cyan.value,
            'y': color.yellow.value,
            'm': color.magenta.value,
            'k': color.black.value,
            'w': color.white.value
        }

        # Create a new independent scene
        self.scene = canvas()

        # Apply the settings
        self.scene.background = color.white
        self.scene.width = width
        self.scene.height = height
        self.scene.autoscale = False

        # Disable default controls
        self.scene.userpan = True  # Keep shift+mouse panning (key overwritten)
        self.scene.userzoom = True  # Keep zoom controls (scrollwheel)
        self.scene.userspin = False  # Remove ctrl+mouse enabled to rotate

        self.__grid_visibility = grid
        self.__camera_lock = False
        self.__grid_relative = True

        # Apply HTML title/caption
        if title != '':
            self.scene.title = title

        self.__default_caption = caption
        if caption != '':
            self.scene.caption = caption

        self.__ui_controls = UIMMap()
        self.__reload_caption()

        # Any time a key or mouse is held down, run the callback function
        rate(30)  # 30Hz
        self.scene.bind('keydown', self.__handle_keyboard_inputs)

        # Create the grid, and display if wanted
        self.__graphics_grid = GraphicsGrid(self.scene, colour=g_col)
        # Toggle grid to 2D
        # self.__graphics_grid.toggle_2d_3d()
        # Lock the grid
        # self.__graphics_grid.set_relative(False)
        # Turn off grid if applicable
        if not self.__grid_visibility:
            self.__graphics_grid.set_visibility(False)

        # Reset the camera to known spot
        self.__reset_camera()
        self.__graphics_grid.update_grid()
        self.__graphics_grid.toggle_2d_3d()

    #######################################
    #  Canvas Management
    #######################################
    def clear_scene(self):
        """
        Clear the scene of all objects, keeping the grid visible if set on
        """
        # Save grid visibility
        restore = self.__grid_visibility

        # Set invis
        if restore:
            self.__graphics_grid.set_visibility(False)

        # Set all objects invis
        for obj in self.scene.objects:
            obj.visible = False

        # Restore grid (if needed)
        if restore:
            self.__graphics_grid.set_visibility(True)

    def grid_visibility(self, is_visible):
        """
        Update the grid visibility in the scene

        :param is_visible: Whether the grid should be visible or not
        :type is_visible: `bool`
        """
        self.__graphics_grid.set_visibility(is_visible)

    #######################################
    #  UI Management
    #######################################
    def __handle_keyboard_inputs(self):
        """
        Pans amount dependent on distance between camera and focus point.
        Closer = smaller pan amount

        A = move left (pan)
        D = move right (pan)
        W = move up (pan)
        S = move down (pan)

        <- = rotate left along camera axes (rotate)
        -> = rotate right along camera axes (rotate)
        ^ = rotate up along camera axes (rotate)
        V = rotate down along camera axes (rotate)

        Q = roll left (rotate)
        E = roll right (rotate)

        shift + LMB = pan (Default Vpython)
        """
        # If camera lock, just skip the function
        if self.__camera_lock:
            return

        # Constants
        pan_amount = 0.02  # units
        rot_amount = 1.0  # deg

        # Current settings
        cam_distance = self.scene.camera.axis.mag
        cam_pos = vector(self.scene.camera.pos)
        cam_focus = vector(self.scene.center)

        # Weird manipulation to get correct vector directions.
        # (scene.camera.up always defaults to world up)
        cam_axis = (vector(self.scene.camera.axis))  # X
        cam_side_axis = self.scene.camera.up.cross(cam_axis)  # Y
        cam_up = cam_axis.cross(cam_side_axis)  # Z

        cam_up.mag = cam_axis.mag

        # Get a list of keys
        keys = keysdown()

        # Userpan uses ctrl, so skip this check to avoid changing camera pose
        # while shift is held
        if 'shift' in keys:
            return

        #####################################################################
        # PANNING
        # Check if the keys are pressed, update vectors as required
        # Changing camera position updates the scene center to follow
        # same changes
        if 'w' in keys:
            cam_pos = cam_pos + cam_up * pan_amount
        if 's' in keys:
            cam_pos = cam_pos - cam_up * pan_amount
        if 'a' in keys:
            cam_pos = cam_pos + cam_side_axis * pan_amount
        if 'd' in keys:
            cam_pos = cam_pos - cam_side_axis * pan_amount

        # Update camera position before rotation
        # (to keep pan and rotate separate)
        self.scene.camera.pos = cam_pos

        #####################################################################
        # Camera Roll
        # If only one rotation key is pressed
        if 'q' in keys and 'e' not in keys:
            # Rotate camera up
            cam_up = cam_up.rotate(angle=-radians(rot_amount), axis=cam_axis)
            # Set magnitude as it went to inf
            cam_up.mag = cam_axis.mag
            # Set
            self.scene.up = cam_up

        # If only one rotation key is pressed
        if 'e' in keys and 'q' not in keys:
            # Rotate camera up
            cam_up = cam_up.rotate(angle=radians(rot_amount), axis=cam_axis)
            # Set magnitude as it went to inf
            cam_up.mag = cam_axis.mag
            # Set
            self.scene.up = cam_up

        ######################################################################
        # CAMERA ROTATION
        d = cam_distance
        move_dist = sqrt(d ** 2 + d ** 2 - 2 * d * d * cos(
            radians(rot_amount)))  # SAS Cosine

        # If only left not right key
        if 'left' in keys and 'right' not in keys:
            # Calculate distance to translate
            cam_pos = cam_pos + norm(cam_side_axis) * move_dist
            # Calculate new camera axis
            cam_axis = -(cam_pos - cam_focus)
        if 'right' in keys and 'left' not in keys:
            cam_pos = cam_pos - norm(cam_side_axis) * move_dist
            cam_axis = -(cam_pos - cam_focus)
        if 'up' in keys and 'down' not in keys:
            cam_pos = cam_pos + norm(cam_up) * move_dist
            cam_axis = -(cam_pos - cam_focus)
        if 'down' in keys and 'up' not in keys:
            cam_pos = cam_pos - norm(cam_up) * move_dist
            cam_axis = -(cam_pos - cam_focus)

        # Update camera position and axis
        self.scene.camera.pos = cam_pos
        self.scene.camera.axis = cam_axis

    def __reset_camera(self):
        """
        Reset the camera to a known position
        """
        # Reset Camera
        self.scene.camera.pos = vector(5, 5, 12)  # Hover above (5, 5, 0)
        # Ever so slightly off focus, to ensure grid is rendered in the right
        # region
        # (if directly at, draws numbers wrong spots)
        # Focus on (5, 5, 0)
        self.scene.camera.axis = vector(-0.001, -0.001, -12)
        self.scene.up = y_axis_vector

    def __reload_caption(self):
        """
        Reload the UI with the new list of robot names
        """
        # Remove all UI elements
        for item in self.__ui_controls:
            if self.__ui_controls.get(item) is None:
                continue
            self.__ui_controls.get(item).delete()
        # Restore the caption
        self.scene.caption = self.__default_caption
        # Create the updated caption.
        self.__setup_ui_controls()

    def __setup_ui_controls(self):
        """
        The initial configuration of the user interface
        """
        self.scene.append_to_caption('\n')

        # Button to reset camera
        btn_reset = button(
            bind=self.__reset_camera, text="Reset Camera")
        self.__ui_controls.btn_reset = btn_reset
        self.scene.append_to_caption('\t')

        chkbox_cam = checkbox(
            bind=self.__camera_lock_checkbox,
            text="Camera Lock", checked=self.__camera_lock)
        self.__ui_controls.chkbox_cam = chkbox_cam
        self.scene.append_to_caption('\t')

        chkbox_rel = checkbox(
            bind=self.__grid_relative_checkbox,
            text="Grid Relative", checked=self.__grid_relative)
        self.__ui_controls.chkbox_rel = chkbox_rel
        self.scene.append_to_caption('\n\n')

        # Button to clear the screen
        btn_clr = button(bind=self.clear_scene, text="Clear Scene")
        self.__ui_controls.btn_clear = btn_clr
        self.scene.append_to_caption('\n\n')

        # Checkbox for grid visibility
        chkbox_grid = checkbox(
            bind=self.__grid_visibility_checkbox, text="Grid Visibility",
            checked=self.__grid_visibility)
        self.__ui_controls.chkbox_grid = chkbox_grid
        self.scene.append_to_caption('\t')

        # Prevent the space bar from toggling the active checkbox/button/etc
        # (default browser behaviour)
        self.scene.append_to_caption('''
                       <script type="text/javascript">
                           $(document).keyup(function(event) {
                               if(event.which === 32) {
                                   event.preventDefault();
                               }
                           });
                       </script>''')
        # https://stackoverflow.com/questions/22280139/prevent-space-button-from-triggering-any-other-button-click-in-jquery

        # Control manual
        controls_str = '<br><b>Controls</b><br>' \
                       '<b>PAN</b><br>' \
                       'SHFT + LMB | <i>free pan</i><br>' \
                       'W , S | <i>up / down</i><br>' \
                       'A , D | <i>left / right</i><br>' \
                       '<b>ROTATE</b><br>' \
                       'ARROWS KEYS | <i>rotate direction</i><br>' \
                       'Q , E | <i>roll left / right</i><br>' \
                       '<b>ZOOM</b></br>' \
                       'MOUSEWHEEL | <i>zoom in / out</i><br>' \
                       '<script type="text/javascript">var arrow_keys_handler = function(e) {switch(e.keyCode){ case 37: case 39: case 38:  case 40: case 32: e.preventDefault(); break; default: break;}};window.addEventListener("keydown", arrow_keys_handler, false);</script>'  # noqa
        # Disable the arrow keys from scrolling in the browser
        # https://stackoverflow.com/questions/8916620/disable-arrow-key-scrolling-in-users-browser
        self.scene.append_to_caption(controls_str)

    #######################################
    # UI CALLBACKS
    #######################################
    def __camera_lock_checkbox(self, c):
        """
        When a checkbox is changed for the camera lock, update the camera

        :param c: The checkbox that has been toggled
        :type c: class:`checkbox`
        """
        # Update parameters
        # True = locked
        self.__camera_lock = c.checked
        # True = enabled
        self.scene.userspin = not c.checked
        self.scene.userzoom = not c.checked

    def __grid_relative_checkbox(self, c):
        """
        When a checkbox is changed for the grid lock, update the grid

        :param c: The checkbox that has been toggled
        :type c: class:`checkbox`
        """
        self.__graphics_grid.set_relative(c.checked)
        self.__grid_relative = c.checked

    def __grid_visibility_checkbox(self, c):
        """
        When a checkbox is changed for the grid visibility, update the graphics

        :param c: The checkbox that has been toggled
        :type c: class:`checkbox`
        """
        self.grid_visibility(c.checked)
        self.__grid_visibility = c.checked

    #######################################
    #  Drawing Functions
    #######################################
    def __draw_path(
            self, x_path, y_path, opt_line, opt_marker,
            opt_colour, thickness=0.05):
        """
        Draw a line from point to point in the 2D path

        :param x_path: The x path to draw on the canvas
        :type x_path: `list`
        :param y_path: The y path to draw on the canvas
        :type y_path: `list`
        :param opt_line: The line option argument
        :type opt_line: `str`
        :param opt_marker: The marker option argument
        :type opt_marker: `str`
        :param opt_colour: The colour option argument
        :type opt_colour: `str`
        :param thickness: Thickness of the line
        :type thickness: `float`
        :raises ValueError: Invalid line type given
        """
        # Get colour
        colour = self.__get_colour_from_string(opt_colour)

        # For every point in the list, draw a line to the next one
        # (excluding last point)
        for point in range(0, len(x_path)):
            # Get point 1
            x1 = x_path[point]
            y1 = y_path[point]
            p1 = vector(x1, y1, 0)

            # If at end / only coordinate - draw a marker
            if point == len(x_path) - 1:
                create_marker(self.scene, x1, y1, opt_marker, colour)
                return

            # Get point 2
            x2 = x_path[point + 1]
            y2 = y_path[point + 1]
            p2 = vector(x2, y2, 0)

            if opt_line == '':
                # Only one marker to avoid double-ups
                create_marker(self.scene, x1, y1, opt_marker, colour)
            elif opt_line == '-':
                create_line(
                    p1, p2, self.scene, colour=colour, thickness=thickness)
                # Only one marker to avoid double-ups
                create_marker(self.scene, x1, y1, opt_marker, colour)
            elif opt_line == '--':
                create_segmented_line(
                    p1, p2, self.scene, 0.3, colour=colour,
                    thickness=thickness)
                # Only one marker to avoid double-ups
                create_marker(self.scene, x1, y1, opt_marker, colour)
            elif opt_line == ':':
                create_segmented_line(
                    p1, p2, self.scene, 0.05, colour=colour,
                    thickness=thickness)
                # Only one marker to avoid double-ups
                create_marker(self.scene, x1, y1, opt_marker, colour)
            elif opt_line == '-.':
                raise NotImplementedError("Other line types not implemented")
            else:
                raise ValueError("Invalid line type given")

    def plot(self, x_coords, y_coords=None, options=''):
        """
        Same usage as MATLAB's plot.

        If given one list of coordinates, plots against index
        If given two lists of coordinates, plots both (1st = x, 2nd = y)

        Options string is identical to MATLAB's input string
        If you do not specify a marker type, plot uses no marker.
        If you do not specify a line style, plot uses a solid line.

            b     blue          .     point              -     solid
            g     green         o     circle             :     dotted
            r     red           x     x-mark             -.    dashdot
            c     cyan          +     plus               --    dashed
            m     magenta       *     star             (none)  no line
            y     yellow        s     square
            k     black         d     diamond
            w     white         v     triangle (down)
                                ^     triangle (up)
                                <     triangle (left)
                                >     triangle (right)
                                p     pentagram
                                h     hexagram

        :param x_coords: The first plane of coordinates to plot
        :type x_coords: `list`
        :param y_coords: The second plane of coordinates to plot with.
        :type y_coords: `list`, `str`, optional
        :param options: A string of options to plot with
        :type options: `str`, optional
        :raises ValueError: Number of X and Y coordinates must be equal
        """
        # TODO
        #  add options for line width, marker size

        # If y-vector is str, then only x vector given
        if isinstance(y_coords, str):
            options = y_coords
            y_coords = None

        one_set_data = False
        # Set y-vector to default if None
        if y_coords is None:
            one_set_data = True
            y_coords = [*range(0, len(x_coords))]

        # Verify x, y coords have same length
        if len(x_coords) != len(y_coords):
            raise ValueError(
                "Number of X coordinates does not equal "
                "number of Y coordinates.")

        # Verify options given (and save settings to be applied)
        verified_options = self.__verify_plot_options(options)

        if one_set_data:
            # Draw plot for one list of data
            self.__draw_path(
                y_coords,  # Y is default x-coords in one data set
                x_coords,  # User input
                verified_options[0],  # Line
                verified_options[1],  # Marker
                verified_options[2],  # Colour
            )
        else:
            # Draw plot for two lists of data
            self.__draw_path(
                x_coords,  # User input
                y_coords,  # User input
                verified_options[0],  # Line
                verified_options[1],  # Marker
                verified_options[2],  # Colour
            )

    def __verify_plot_options(self, options_str):
        """
        Verify that the given options are usable.

        :param options_str: The given options from the plot command to verify
        user input

        :type options_str: `str`
        :raises ValueError: Unknown character entered
        :raises ValueError: Too many line segments used
        :raises ValueError: Too many marker segments used
        :raises ValueError: Too many colour segments used
        :returns: List of options to plot with
        :rtype: `list`
        """
        default_line = '-'
        default_marker = ''
        default_colour = 'k'

        # Split str into chars list
        options_split = list(options_str)

        # If 0, set defaults and return early
        if len(options_split) == 0:
            return [default_line, default_marker, default_colour]

        # If line_style given, join the first two options if applicable
        # (some types have 2 characters)
        for char in range(0, len(options_split) - 1):
            # If char is '-' (only leading character in double length option)
            if options_split[char] == '-' and len(options_split) > 1:
                # If one of the leading characters is valid
                if options_split[char + 1] == '-' or \
                        options_split[char + 1] == '.':
                    # Join the two into the first
                    options_split[char] = options_split[char] \
                                          + options_split[char + 1]
                    # Shuffle down the rest
                    for idx in range(char + 2, len(options_split)):
                        options_split[idx - 1] = options_split[idx]
                    # Remove duplicate extra
                    options_split.pop()

        # If any unknown, throw error
        for option in options_split:
            if option not in self.__line_styles and \
                    option not in self.__marker_styles and \
                    option not in self.__colour_styles:
                error_string = "Unknown character entered: '{0}'"
                raise ValueError(error_string.format(option))

        ##############################
        # Verify Line Style
        ##############################
        line_style_count = 0  # Count of options used
        # Index position of index used (only used when count == 1)
        line_style_index = 0
        for option in options_split:
            if option in self.__line_styles:
                line_style_count = line_style_count + 1
                line_style_index = self.__line_styles.index(option)

        # If more than one, throw error
        if line_style_count > 1:
            raise ValueError(
                "Too many line style arguments given. Only one allowed")
        # If none, set as solid
        elif line_style_count == 0 or not any(
                item in options_split for item in self.__line_styles):
            output_line = default_line
        # If one, set as given
        else:
            output_line = self.__line_styles[line_style_index]
        ##############################

        ##############################
        # Verify Marker Style
        ##############################
        marker_style_count = 0  # Count of options used
        # Index position of index used (only used when count == 1)
        marker_style_index = 0
        for option in options_split:
            if option in self.__marker_styles:
                marker_style_count = marker_style_count + 1
                marker_style_index = self.__marker_styles.index(option)

        # If more than one, throw error
        if marker_style_count > 1:
            raise ValueError(
                "Too many marker style arguments given. Only one allowed")
        # If none, set as no-marker
        elif marker_style_count == 0 or not any(
                item in options_split for item in self.__marker_styles):
            output_marker = default_marker
        # If one, set as given
        else:
            output_marker = self.__marker_styles[marker_style_index]
            # If marker set and no line given, turn line to no-line
            if line_style_count == 0 or not any(
                    item in options_split for item in self.__line_styles):
                output_line = ''
        ##############################

        ##############################
        # Verify Colour Style
        ##############################
        colour_style_count = 0  # Count of options used
        # Index position of index used (only used when count == 1)
        colour_style_index = 0
        for option in options_split:
            if option in self.__colour_styles:
                colour_style_count = colour_style_count + 1
                colour_style_index = self.__colour_styles.index(option)

        # If more than one, throw error
        if colour_style_count > 1:
            raise ValueError(
                "Too many colour style arguments given. Only one allowed")
        # If none, set as black
        elif colour_style_count == 0 or not any(
                item in options_split for item in self.__colour_styles):
            output_colour = default_colour
        # If one, set as given
        else:
            output_colour = self.__colour_styles[colour_style_index]
        ##############################

        return [output_line, output_marker, output_colour]

    def __get_colour_from_string(self, colour_string):
        """
        Using the colour plot string input, return an rgb array of the colour
        selected

        :param colour_string: The colour string option
        :type colour_string: `str`
        :returns: List of RGB values for the representative colour
        :rtype: `list`
        """
        # Return the RGB list (black if not in dictionary)
        return self.__colour_dictionary.get(colour_string, color.black.value)


def convert_grid_to_z_up(scene):  # pragma nocover
    """
    Rotate the camera so that +z is up
    (Default vpython scene is +y up)
    """

    '''
    There is an interaction between up and forward, the direction that the
    camera is pointing. By default, the camera points in the -z direction
    vector(0,0,-1). In this case, you can make the x or y axes (or anything
    between) be the up vector, but you cannot make the z axis be the up
    vector, because this is the axis about which the camera rotates when
    you set the up attribute. If you want the z axis to point up, first set
    forward to something other than the -z axis, for example vector(1,0,0).
    https://www.glowscript.org/docs/VPythonDocs/canvas.html
    '''
    # First set the x-axis forward
    scene.forward = x_axis_vector
    scene.up = z_axis_vector

    # Place the camera in the + axes
    scene.camera.pos = vector(10, 10, 10)
    scene.camera.axis = -scene.camera.pos
    return


def draw_reference_frame_axes(se3_pose, scene):  # pragma nocover
    """
    Draw x, y, z axes from the given point.
    Each axis is represented in the objects reference frame.


    :param se3_pose: SE3 pose representation of the reference frame
    :type se3_pose: class:`spatialmath.pose3d.SE3`
    :param scene: Which scene to put the graphics in
    :type scene: class:`vpython.canvas`
    :return: Compound object of the 3 axis arrows.
    :rtype: class:`vpython.compound`
    """

    origin = get_pose_pos(se3_pose)
    x_axis = get_pose_x_vec(se3_pose)
    y_axis = get_pose_y_vec(se3_pose)

    # Create Basic Frame
    # Draw X Axis
    x_arrow = arrow(
        canvas=scene, pos=origin, axis=x_axis_vector,
        length=0.25, color=color.red)

    # Draw Y Axis
    y_arrow = arrow(
        canvas=scene, pos=origin, axis=y_axis_vector,
        length=0.25, color=color.green)

    # Draw Z Axis
    z_arrow = arrow(
        canvas=scene, pos=origin, axis=z_axis_vector,
        length=0.25, color=color.blue)

    # Combine all to manipulate together
    # Set origin to where axis converge (instead of the middle of
    # the resulting object bounding box)
    frame_ref = compound(
        [x_arrow, y_arrow, z_arrow], origin=origin, canvas=scene)

    # Set frame axes
    frame_ref.axis = x_axis
    frame_ref.up = y_axis

    return frame_ref



#!/usr/bin/env python

import numpy as np
from roboticstoolbox.robot.ETS import ETS
from roboticstoolbox.robot.ERobot import ERobot
from roboticstoolbox.robot.ELink import ELink


class Panda(ERobot):
    """
    Create model of Franka-Emika Panda manipulator

    panda = Panda() creates a robot object representing the Franka-Emika
    Panda robot arm. This robot is represented using the elementary
    transform sequence (ETS).

    ETS taken from [1] based on
    https://frankaemika.github.io/docs/control_parameters.html

    :references:
        - Kinematic Derivatives using the Elementary Transform
          Sequence, J. Haviland and P. Corke

    """
    def __init__(self):

        deg = np.pi/180
        mm = 1e-3
        tool_offset = (103)*mm

        l0 = ELink(
            ETS.tz(0.333) * ETS.rz(),
            name='link0',
            parent=None
        )

        l1 = ELink(
            ETS.rx(-90*deg) * ETS.rz(),
            name='link1',
            parent=l0
        )

        l2 = ELink(
            ETS.rx(90*deg) * ETS.tz(0.316) * ETS.rz(),
            name='link2',
            parent=l1
        )

        l3 = ELink(
            ETS.tx(0.0825) * ETS.rx(90, 'deg') * ETS.rz(),
            name='link3',
            parent=l2
        )

        l4 = ELink(
            ETS.tx(-0.0825) * ETS.rx(-90, 'deg') * ETS.tz(0.384) * ETS.rz(),
            name='link4',
            parent=l3
        )

        l5 = ELink(
            ETS.rx(90, 'deg') * ETS.rz(),
            name='link5',
            parent=l4
        )

        l6 = ELink(
            ETS.tx(0.088) * ETS.rx(90, 'deg') * ETS.tz(0.107) * ETS.rz(),
            name='link6',
            parent=l5
        )

        ee = ELink(
            ETS.tz(tool_offset) * ETS.rz(-np.pi/4),
            name='ee',
            parent=l6
        )

        elinks = [l0, l1, l2, l3, l4, l5, l6, ee]

        super(Panda, self).__init__(
            elinks,
            name='Panda',
            manufacturer='Franka Emika')

        self.addconfiguration(
            "qz", np.array([0, 0, 0, 0, 0, 0, 0]))
        self.addconfiguration(
            "qr", np.array([0, -0.3, 0, -2.2, 0, 2.0, np.pi/4]))


if __name__ == '__main__':   # pragma nocover

    robot = Panda()
    print(robot)

    print(robot.n, robot.M)
    print(robot.elinks)
    print(robot.q_idx)
    for link in robot:
        print(link.name, link)
        print(link.ets, link.v, link.isjoint)
        print(link.parent, link.child)
        print(link.Ts)
        print(link.geometry, link.collision)

        print()

    from ansitable import ANSITable, Column

    table = ANSITable(
        Column("link"),
        Column("parent"),
        Column("ETS", headalign="^", colalign="<"),
        border="thin")
    for link in robot:
        if link.isjoint:
            color = ""
        else:
            color = "<<blue>>"
        table.row(
            color + link.name,
            link.parent.name if link.parent is not None else "-",
            link.ets * link.v if link.v is not None else link.ets)
    table.print()



#!/usr/bin/env python
"""
@author Micah Huth
"""

import numpy as np
from vpython import box, compound, color, sqrt
from roboticstoolbox.backends.VPython.canvas import draw_reference_frame_axes
from roboticstoolbox.backends.VPython.common_functions import \
    array, x_axis_vector, y_axis_vector, z_axis_vector, \
    get_pose_pos, get_pose_x_vec, get_pose_y_vec, get_pose_z_vec, \
    vector
from roboticstoolbox.backends.VPython.stl import set_stl_origin, \
    import_object_from_numpy_stl
from time import perf_counter
from spatialmath import SE3
from pathlib import PurePath


class DefaultJoint:  # pragma nocover
    """
    This class forms a base for all the different types of joints
    - Rotational
    - Prismatic
    - Static
    - Gripper

    :param initial_se3: Pose to set the joint to initially
    :type initial_se3: class:`spatialmath.pose3d.SE3`
    :param structure: A variable representing the joint length (float) or
        a file path to an STL (str)
    :type structure: `float`, `str`
    :param g_canvas: The canvas in which to add the link
    :type g_canvas: class:`graphics.graphics_canvas.graphicscanvas3d`
    :param qlim: A list of the angle limits for the joint
    :type qlim: `list`
    :param theta: The current angle of the joint in radians
    :type theta: `float`
    :param axis_through: The axis that the longest side goes through
    :type axis_through: class:`numpy.ndarray`
    """

    def __init__(
            self, initial_se3, structure, g_canvas,
            qlim, theta, axis_through=array([1, 0, 0])):

        if not isinstance(
                structure, float) and not isinstance(structure, PurePath):
            error_str = "structure must be of type {0} or {1}. Given {2}. " \
                "Either give a length (float)," \
                "or meshdata [filepath, scale, origindata]"
            raise TypeError(error_str.format(float, str, type(structure)))

        self.qlim = qlim
        self.theta = theta

        self.__scene = g_canvas.scene
        self.__pose = initial_se3

        # Set the graphic
        # self.stl_offset = None
        self.__graphic_obj = self.__set_graphic(structure, axis_through)
        self.visible = True

        # Set the other reference frame vectors
        self.__graphic_ref = draw_reference_frame_axes(
            self.__pose, self.__scene)

        # Apply the initial pose
        self.update_pose(self.__pose)

    def update_position(self, se_object):
        """
        Given an SE object, update just the orientation of the joint.

        :param se_object: SE3 pose representation of the joint
        :type se_object: class:`spatialmath.pose3d.SE3`
        """
        new_position = get_pose_pos(se_object)
        self.__graphic_obj.pos = new_position

        # Update the reference frame
        self.__update_reference_frame()
        self.draw_reference_frame(self.__graphic_ref.visible)

    def update_orientation(self, se_object):
        """
        Given an SE object, update just the orientation of the joint.

        :param se_object: SE3 pose representation of the joint
        :type se_object: class:`spatialmath.pose3d.SE3`
        """
        # Get the new pose details
        new_x_axis = get_pose_x_vec(se_object)
        new_y_axis = get_pose_y_vec(se_object)
        # new_z_axis = get_pose_z_vec(se_object)  # not needed

        # Update the graphic object
        self.__graphic_obj.axis = new_x_axis
        self.__graphic_obj.up = new_y_axis

        # Update the reference frame
        self.__update_reference_frame()
        self.draw_reference_frame(self.__graphic_ref.visible)

    def update_pose(self, se_object):
        """
        Given an SE object, update the pose of the joint.

        :param se_object: SE3 pose representation of the joint
        :type se_object: class:`spatialmath.pose3d.SE3`
        """
        # if self.stl_offset is not None:
        #     calc_se3 = se_object * self.stl_offset
        # else:
        #     calc_se3 = se_object

        # Get the new pose details
        new_x_axis = get_pose_x_vec(se_object)
        new_y_axis = get_pose_y_vec(se_object)
        # new_z_axis = get_pose_z_vec(se_object)  # not needed
        new_position = get_pose_pos(se_object)

        # Update the graphic object
        self.__graphic_obj.pos = new_position
        self.__graphic_obj.axis = new_x_axis
        self.__graphic_obj.up = new_y_axis

        # Update the reference frame
        self.__pose = se_object
        self.__update_reference_frame()
        self.draw_reference_frame(self.__graphic_ref.visible)

    def __update_reference_frame(self):
        """
        Update the reference frame axis vectors
        """
        self.__x_vector = get_pose_x_vec(self.__pose)
        self.__y_vector = get_pose_y_vec(self.__pose)
        self.__z_vector = get_pose_z_vec(self.__pose)

    def draw_reference_frame(self, is_visible):
        """
        Draw a reference frame at the tool point position.

        :param is_visible: Whether the reference frame should be drawn or not
        :type is_visible: `bool`
        """
        self.__graphic_ref.pos = get_pose_pos(self.__pose)
        self.__graphic_ref.axis = get_pose_x_vec(self.__pose)
        self.__graphic_ref.up = get_pose_y_vec(self.__pose)
        self.__graphic_ref.visible = is_visible

    def set_stl_joint_origin(self, current_location, required_location):
        """
        Modify the origin position of the graphical object.
        This is mainly used when loading STL objects. If the origin (point of
        rotation/placement) does not align with
        reality, it can be set.
        It translates the object to place the origin at the required location,
        and save the new origin to the object.

        :param current_location: 3D coordinate of where the real origin is in
            space.
        :type current_location: class:`vpython.vector`
        :param required_location: 3D coordinate of where the real origin
            should be in space
        :type required_location: class:`vpython.vector`
        """
        set_stl_origin(
            self.__graphic_obj, current_location,
            required_location, self.__scene)

    def set_joint_visibility(self, is_visible):
        """
        Choose whether or not the joint is displayed in the canvas.

        :param is_visible: Whether the joint should be drawn or not
        :type is_visible: `bool`
        """
        # If the option is different to current setting
        if is_visible is not self.visible:
            # Update
            self.__graphic_obj.visible = is_visible
            # self.__graphic_ref.visible = is_visible
            self.visible = is_visible

    def __set_graphic(self, structure, axis_through):
        """
        Set the graphic object depending on if one was given. If no object was
        given, create a box and return it

        :param structure: `float` or `str` representing the joint length or
            STL path to load from
        :type structure: `float`, `str`
        :param axis_through: The axis that the longest side goes through
        :type axis_through: class:`numpy.ndarray`
        :raises ValueError: Joint length must be greater than 0
        :return: Graphical object for the joint
        :rtype: class:`vpython.compound`
        """
        if isinstance(structure, float):
            length = structure
            if length <= 0.0:
                raise ValueError("Joint length must be greater than 0")
            axis = vector(axis_through[0], axis_through[1], axis_through[2])
            axis.mag = length

            box_midpoint = axis / 2
            box_tooltip = axis

            # Create a box along the +x axis, with the origin
            # (point of rotation) at (0, 0, 0)
            graphic_obj = box(
                canvas=self.__scene,
                pos=vector(box_midpoint.x, box_midpoint.y, box_midpoint.z),
                axis=axis,
                size=vector(length, 0.1, 0.1),  # L, W, H
                # up=y_axis_vector
            )

            # Set the boxes new origin
            graphic_obj = compound(
                [graphic_obj], origin=box_tooltip, axis=axis)

            return graphic_obj
        else:
            # self.stl_offset = structure[2]
            return import_object_from_numpy_stl(structure, self.__scene)

    def set_texture(self, colour=None, texture_link=None):
        """
        Apply the texture/colour to the object. If both are given, both are
        applied.
        Texture link can either be a link to an online image, or local file.

        WARNING: If the texture can't be loaded, the object will have n
        texture
        (appear invisible, but not set as invisible).

        WARNING: If the image has a width or height that is not a power of 2
        (that is, not 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, etc.),
        the image is stretched to the next larger width or height that is
        a power of 2.

        :param colour: List of RGB values
        :type colour: `list`, optional
        :param texture_link: Path/website to a texture image
        :type texture_link: `str`, optional
        :raises ValueError: RGB values must be normalised between 0 and 1
        """
        # Apply the texture
        if texture_link is not None:
            self.__graphic_obj.texture = {
                'file': texture_link
                # 'bumpmap', 'place', 'flipx', 'flipy', 'turn'
            }
            # Wait for the texture to load
            self.__scene.waitfor("textures")
        else:
            # Remove any texture
            self.__graphic_obj.texture = None

        # Apply the colour
        if colour is not None:
            if colour[0] > 1.0 or colour[1] > 1.0 or colour[2] > 1.0 or \
               colour[0] < 0.0 or colour[1] < 0.0 or colour[2] < 0.0:
                raise ValueError(
                    "RGB values must be normalised between 0 and 1")
            colour_vec = vector(colour[0], colour[1], colour[2])
            self.__graphic_obj.color = colour_vec
        else:
            # Set to white if none given
            self.__graphic_obj.color = color.white

    def set_transparency(self, opacity):
        """
        Sets the transparency of the joint.

        :param opacity: Normalised value (0 -> 1) to set the opacity.
            0 = transparent, 1 = opaque
        :type opacity: `float`
        :raises ValueError: Value must be between 0 and 1 inclusively
        """
        if opacity < 0 or opacity > 1:
            raise ValueError("Value must be between 0 and 1 inclusively")
        self.__graphic_obj.opacity = opacity

    def get_axis_vector(self, axis):
        """
        Get the current vector of a specified X, Y, or Z axis

        :param axis: Specified joint axis to get the angle of rotation of
        :type axis: class:`vpython.vector`
        :return: Current vector representation of the joints X, Y, or Z axis
        :rtype: class:`vpython.vector`
        :raise ValueError: The given axis_of_rotation must be one of the
            default X, Y, Z vectors (e.g. x_axis_vector)
        """
        # Return new vectors to avoid pass by reference
        if axis.equals(x_axis_vector):
            return self.__x_vector
        elif axis.equals(y_axis_vector):
            return self.__y_vector
        elif axis.equals(z_axis_vector):
            return self.__z_vector
        else:
            error_str = "Bad input vector given ({0}). " \
                        "Must be either x_axis_vector ({1}), " \
                        "y_axis_vector ({2})," \
                        "or z_axis_vector ({3})."
            raise ValueError(
                error_str.format(
                    axis, x_axis_vector, y_axis_vector, z_axis_vector))

    def get_pose(self):
        """
        Return the current pose of the joint

        :return: SE3 representation of the current joint pose
        :rtype: class:`spatialmath.pose3d.SE3`
        """
        return self.__pose

    def get_joint_type(self):
        """
        Return the type of joint (To Be Overridden by Child classes)

        :return: String representation of the joint type.
        :rtype: `str`
        """
        return ""

    def get_graphic_object(self):
        """
        Getter function that returns the graphical object of the joint

        :return: VPython graphical entity of the joint
        :rtype: class:`vpython.object`
        """
        return self.__graphic_obj

    def get_scene(self):
        """
        Getter function that returns the scene the object is in

        :return: The scene the object is in
        :rtype: class:`vpython.canvas`
        """
        return self.__scene


class RotationalJoint(DefaultJoint):  # pragma nocover
    """
    A rotational joint based off the default joint class

    :param g_canvas: The canvas in which to add the link
    :type g_canvas: class:`graphics.graphics_canvas.graphicscanvas3d`
    :param initial_se3: Pose to set the joint to initially
    :type initial_se3: class:`spatialmath.pose3d.SE3`
    :param structure: either a float of the length of the joint, or a list of
        str of the filepath and scale
    :type structure: `float`, `list`
    :param qlim: A list of the angle limits for the joint
    :type qlim: `list`
    :param theta: The current angle of the joint in radians
    :type theta: `float`
    :param axis_through: The axis that the longest side goes through
    :type axis_through: class:`numpy.ndarray`
    """

    def __init__(
            self, initial_se3, structure, g_canvas, qlim,
            theta, axis_through=array([1, 0, 0])):

        # Call super init function
        super().__init__(
            initial_se3, structure, g_canvas, qlim, theta, axis_through)

    def get_joint_type(self):
        """
        Return the type of joint (R for Rotational)

        :return: String representation of the joint type.
        :rtype: `str`
        """
        return "R"


class PrismaticJoint(DefaultJoint):  # pragma nocover
    """
    A prismatic joint based from the default joint class

    :param g_canvas: The canvas in which to add the link
    :type g_canvas: class:`graphics.graphics_canvas.graphicscanvas3d`
    :param initial_se3: Pose to set the joint to initially
    :type initial_se3: class:`spatialmath.pose3d.SE3`
    :param structure: either a float of the length of the joint, or a list of
        str of the filepath and scale
    :type structure: `float`, `list`
    :param qlim: A list of the angle limits for the joint
    :type qlim: `list`
    :param theta: The current angle of the joint in radians
    :type theta: `float`
    :param axis_through: The axis that the longest side goes through
    :type axis_through: class:`numpy.ndarray`
    """
    def __init__(
            self, initial_se3, structure, g_canvas, qlim, theta,
            axis_through=array([1, 0, 0])):

        super().__init__(
            initial_se3, structure, g_canvas, qlim, theta, axis_through)
        self.min_translation = None
        self.max_translation = None

    def translate_joint(self, new_translation):
        # TODO
        pass

    def get_joint_type(self):
        """
        Return the type of joint (P for Prismatic)

        :return: String representation of the joint type.
        :rtype: `str`
        """
        return "P"


class StaticJoint(DefaultJoint):  # pragma nocover
    """
    This class represents a static joint (one that doesn't translate or
    rotate on it's own).
    It has no extra functions to utilise.

    :param g_canvas: The canvas in which to add the link
    :type g_canvas: class:`graphics.graphics_canvas.graphicscanvas3d`
    :param initial_se3: Pose to set the joint to initially
    :type initial_se3: class:`spatialmath.pose3d.SE3`
    :param structure: either a float of the length of the joint, or a list of
        str of the filepath and scale
    :type structure: `float`, `list`
    :param qlim: A list of the angle limits for the joint
    :type qlim: `list`
    :param theta: The current angle of the joint in radians
    :type theta: `float`
    :param axis_through: The axis that the longest side goes through
    :type axis_through: class:`numpy.ndarray`
    """

    def __init__(
            self, initial_se3, structure, g_canvas, qlim, theta,
            axis_through=array([1, 0, 0])):

        super().__init__(
            initial_se3, structure, g_canvas, qlim, theta, axis_through)

    def get_joint_type(self):
        """
        Return the type of joint (S for Static)

        :return: String representation of the joint type.
        :rtype: `str`
        """
        return "S"


class Gripper(DefaultJoint):  # pragma nocover
    """
    This class represents a gripper joint with a moving gripper
    (To Be Implemented).
    Usually the end joint of a robot.

    :param g_canvas: The canvas in which to add the link
    :type g_canvas: class:`graphics.graphics_canvas.graphicscanvas3d`
    :param initial_se3: Pose to set the joint to initially
    :type initial_se3: class:`spatialmath.pose3d.SE3`
    :param structure: either a float of the length of the joint, or a list
        of str of the filepath and scale
    :type structure: `float`, `list`
    :param qlim: A list of the angle limits for the joint
    :type qlim: `list`
    :param theta: The current angle of the joint in radians
    :type theta: `float`
    :param axis_through: The axis that the longest side goes through
    :type axis_through: class:`numpy.ndarray`
    """

    def __init__(
            self, initial_se3, structure, g_canvas,
            qlim, theta, axis_through=array([1, 0, 0])):

        super().__init__(
            initial_se3, structure, g_canvas, qlim, theta, axis_through)

    # TODO close/open gripper

    def get_joint_type(self):
        """
        Return the type of joint (G for Gripper)

        :return: String representation of the joint type.
        :rtype: `str`
        """
        return "G"


class GraphicalRobot:  # pragma nocover
    """
    The GraphicalRobot class holds all of the different joints to easily
    control the robot arm.

    :param graphics_canvas: The canvas to add the robot to
    :type graphics_canvas: class:`GraphicsCanvas`
    :param name: The name of the robot to identify it
    :type name: `str`
    :param robot: A serial link object to create a robot on
    :type robot: class:`roboticstoolbox.robot.DHRobot`
    """
    def __init__(self, graphics_canvas, name, robot=None):
        self.joints = []
        self.num_joints = 0
        self.rob_shown = True
        self.ref_shown = True
        self.opacity = 1
        self.name = name
        self.__scene = graphics_canvas

        self.angles = []
        self.robot = robot

        # If Robot given, create the robot
        if self.robot is not None:
            # Update name
            self.name = self.robot.name
            # Get initial poses
            zero_angles = np.zeros((self.robot.n,))
            all_poses = self.robot.fkine_all(zero_angles, old=False)
            # Create the base
            if robot.basemesh is not None:
                self.append_link("s", all_poses[0], robot.basemesh, [0, 0], 0)
            # else: assume no base joint
            robot_colours = robot.linkcolormap()
            # Create the joints
            for i, link in enumerate(self.robot.links):
                # Get info
                if link.isprismatic:
                    j_type = 'p'
                elif link.isrevolute:
                    j_type = 'r'
                else:
                    j_type = 's'
                pose = all_poses[i+1]   # link frame pose
                if link.mesh is None:
                    if i == 0:
                        x1, x2 = SE3().t[0], all_poses[i].t[0]
                        y1, y2 = SE3().t[1], all_poses[i].t[1]
                        z1, z2 = SE3().t[2], all_poses[i].t[2]
                        length = sqrt(
                            (x2 - x1) * (x2 - x1) + (y2 - y1)
                            * (y2 - y1) + (z2 - z1) * (z2 - z1))  # Length
                    else:
                        x1, x2 = all_poses[i - 1].t[0], all_poses[i].t[0]
                        y1, y2 = all_poses[i - 1].t[1], all_poses[i].t[1]
                        z1, z2 = all_poses[i - 1].t[2], all_poses[i].t[2]
                        length = sqrt(
                            (x2 - x1) * (x2 - x1) + (y2 - y1)
                            * (y2 - y1) + (z2 - z1) * (z2 - z1))  # Length
                else:
                    length = link.mesh
                angle_lims = link.qlim  # Angle limits
                theta = link.theta  # Current angle
                self.append_link(j_type, pose, length, angle_lims, theta)

            # Apply colours
            for i, joint in enumerate(self.joints):
                link_colour = list(robot_colours(i))[:3]
                joint.set_texture(colour=link_colour)

        # Add the robot to the canvas UI
        graphics_canvas.add_robot(self)

    def append_made_link(self, joint):
        """
        Append an already made joint to the end of the robot
        (Useful for links manually created)

        :param joint: A joint object already constructed
        :type joint: class:`graphics.graphics_robot.RotationalJoint`,
            class:`graphics.graphics_robot.PrismaticJoint`,
        class:`graphics.graphics_robot.StaticJoint`,
            class:`graphics.graphics_robot.Gripper`
        :raises RuntimeError: Ensure the link is in the same scene as the robot
        """
        if joint.get_scene() != self.__scene.scene:
            raise RuntimeError(
                "The given made link is not in the same scene "
                "as the robot is.")

        self.joints.append(joint)
        self.num_joints += 1
        self.angles.append(joint.theta)

    def append_link(
            self, typeof, pose, structure, qlim, theta,
            axis_through=array([1, 0, 0])):
        """
        Append a joint to the end of the robot.

        :param typeof: String character of the joint type. e.g. 'R', 'P',
            'S', 'G'
        :type typeof: `str`
        :param pose: SE3 object for the pose of the joint
        :type pose: class:`spatialmath.pose3d.SE3`
        :param structure: either a float of the length of the joint, or a list
            of str of the filepath and scale
        :type structure: `float`, `list`
        :param qlim: A list of the angle limits for the joint
        :type qlim: `list`
        :param theta: The current angle of the joint in radians
        :type theta: `float`
        :param axis_through: The axis that the longest side goes through
        :type axis_through: class:`numpy.ndarray`
        :raises ValueError: typeof must be a valid character
        """
        # Capitalise the type for case-insensitive use
        typeof = typeof.upper()

        if typeof == 'R':
            link = RotationalJoint(
                pose, structure, self.__scene, qlim, theta, axis_through)
        elif typeof == 'P':
            link = PrismaticJoint(
                pose, structure, self.__scene, qlim, theta, axis_through)
        elif typeof == 'S':
            link = StaticJoint(
                pose, structure, self.__scene, qlim, theta, axis_through)
        elif typeof == 'G':
            link = Gripper(
                pose, structure, self.__scene, qlim, theta, axis_through)
        else:
            raise ValueError(
                "typeof should be (case-insensitive) either 'R' (Rotational)"
                ", 'P' (Prismatic), 'S' (Static), or 'G' (Gripper)")

        # Append the joint to the robot
        self.joints.append(link)
        self.num_joints += 1
        self.angles.append(theta)

    def detach_link(self):
        """
        Detach the end link of the robot.

        :raises UserWarning: Must have a joint available to detach
        """
        # Check if no joints to detach
        if self.num_joints == 0:
            raise UserWarning("No robot joints to detach")

        # Turn off the graphics in the canvas
        self.joints[self.num_joints - 1].set_joint_visibility(False)
        self.joints[self.num_joints - 1].draw_reference_frame(False)
        # Ensure deletion
        self.joints[self.num_joints - 1] = None

        # Resize lists
        self.joints = self.joints[0:self.num_joints - 1]
        self.angles = self.angles[0:self.num_joints - 1]
        self.num_joints -= 1

    def set_robot_visibility(self, is_visible):
        """
        Set the entire robots visibility inside the canvas.

        :param is_visible: Whether the robot should be visible or not.
        :type is_visible: `bool`
        """
        if is_visible is not self.rob_shown:
            for joint in self.joints:
                joint.set_joint_visibility(is_visible)
                self.rob_shown = is_visible

    def set_reference_visibility(self, is_visible):
        """
        Set the visibility of the reference frame for all joints.

        :param is_visible: Whether the reference frames should be
            visible or not.
        :type is_visible: `bool`
        """
        if is_visible is not self.ref_shown:
            for joint in self.joints:
                joint.draw_reference_frame(is_visible)
            self.ref_shown = is_visible

    def set_transparency(self, opacity):
        """
        Set the transparency of the robot.
        Allows for easier identification of the reference frames
        (if hidden by the robot itself)

        :param opacity: Normalised value (0 -> 1) to set the opacity.
            0 = transparent, 1 = opaque
        :type opacity: `float`
        """
        if opacity is not self.opacity:
            for joint in self.joints:
                joint.set_transparency(opacity)
            self.opacity = opacity

    def set_joint_poses(self, all_poses):
        """
        Set the joint poses.

        :param all_poses: List of all the new poses to set
        :type all_poses: class:`spatialmath.pose3d.SE3` list
        :raises UserWarning: Robot must not have 0 joints, and given poses
            length must equal number of joints.
        """
        # Sanity checks
        if self.num_joints == 0:
            raise UserWarning(
                "Robot has 0 joints. Create some using append_link()")

        # If given a base pose, when there isn't one
        if self.num_joints == len(all_poses) - 1 and all_poses[0] == SE3():
            # Update the joints
            for idx in range(self.num_joints):
                self.joints[idx].update_pose(all_poses[idx+1])
            return

        # If given all poses excluding a base pose
        if self.num_joints - 1 == len(all_poses):
            for idx in range(1, self.num_joints):
                self.joints[idx].update_pose(all_poses[idx-1])
            return

        # If incorrect number of joints
        if self.num_joints != len(all_poses):
            err = "Number of given poses {0} does not equal number " \
                "of joints {1}"
            raise UserWarning(err.format(len(all_poses), self.num_joints))

        # Update the joints
        for idx in range(self.num_joints):
            self.joints[idx].update_pose(all_poses[idx])

    def animate(self, frame_poses, fps):
        """
        Calling this function will animate the robot through its frames.

        :param frame_poses: A 2D list of each joint pose for each frame.
        :type frame_poses: `list`
        :param fps: Number of frames per second to render at
            (limited by number of graphics trying to update)
        :type fps: `int`
        :raises ValueError: Number of frames and fps must be greater than 0
        """
        num_frames = len(frame_poses)
        # Validate num_frames
        if num_frames == 0:
            raise ValueError(
                "0 frames were given. Supply at least 1 iteration of poses.")

        if fps <= 0:
            raise ValueError(
                "fps must be greater than 0.")
        f = 1 / fps

        for poses in frame_poses:
            # Get current time
            t_start = perf_counter()

            self.set_joint_poses(poses)  # Validation done in set_joint_poses
            # Wait for scene to finish drawing
            self.__scene.scene.waitfor("draw_complete")

            # Get current time
            t_stop = perf_counter()

            # Wait for time of frame to finish
            # If drawing takes longer than frame frequency,
            # this while is skipped
            while t_stop - t_start < f:
                t_stop = perf_counter()

    def fkine(self, joint_angles):
        """
        Call fkine for the robot. If it is based on a seriallink object,
        run it's fkine function.

        :param joint_angles: List of the joint angles
        :type joint_angles: `list`
        """
        # If seriallink object, run it's fkine
        if self.robot is not None:
            return self.robot.fkine_all(joint_angles)
        # Else TODO
        else:
            pass

    def print_joint_poses(self):
        """
        Print all of the current joint poses
        """
        # For each joint
        num = 0
        print(self.name)
        for joint in self.joints:
            print("Joint", num, "| Type:", joint.get_joint_type(), "| Pose:")
            print(joint.get_pose(), "\n")
            num += 1



"""
Python Vehicle
@Author: Kristian Gibson
@Author: Peter Corke
"""
from abc import ABC, abstractmethod
import warnings
from math import pi, sin, cos, tan, atan2
import numpy as np
from scipy import integrate, linalg, interpolate

import matplotlib.pyplot as plt
from matplotlib import patches
import matplotlib.transforms as mtransforms

from spatialmath import SE2, base
from roboticstoolbox import loaddata
from roboticstoolbox.mobile.drivers import VehicleDriver


class Vehicle(ABC):
    def __init__(self, covar=None, speed_max=np.inf, accel_max=np.inf, x0=None, dt=0.1,
                 control=None, animation=None, verbose=False, dim=10):
        r"""
        Superclass for vehicle kinematic models

        :param covar: odometry covariance, defaults to zero
        :type covar: ndarray(2,2), optional
        :param speed_max: maximum speed, defaults to :math:`\infty`
        :type speed_max: float, optional
        :param accel_max: maximum acceleration, defaults to :math:`\infty`
        :type accel_max: float, optional
        :param x0: Initial state, defaults to (0,0,0)
        :type x0: array_like(3), optional
        :param dt: sample update interval, defaults to 0.1
        :type dt: float, optional
        :param control: vehicle control inputs, defaults to None
        :type control: array_like(2), interp1d, VehicleDriver
        :param animation: Graphical animation of vehicle, defaults to None
        :type animation: VehicleAnimation subclass, optional
        :param verbose: print lots of info, defaults to False
        :type verbose: bool, optional
        :param dim: dimensions of 2D plot area, defaults to (-10:10) x (-10:10),
            see :func:`~spatialmath.base.animate.plotvol2`
        :type dims: float, array_like(2), , array_like(4)

        This is an abstract superclass that simulates the motion of a mobile
        robot under the action of a controller.  The controller provides
        control inputs to the vehicle, and the output odometry is returned.
        The true state, effectively unknowable in practice, is computed
        and accessible.

        :seealso: :func:`Bicycle`, :func:`Unicycle`
        """

        if covar is None:
            covar = np.zeros((2,2))
        self._V = covar
        self._dt = dt
        if x0 is None:
            x0 = np.zeros((3,), dtype=float)
        else:
            x0 = base.getvector(x0)
            if len(x0) not in (2,3):
                raise ValueError('x0 must be length 2 or 3')
        self._x0 = x0
        self._x = x0
        
        self._speed_max = speed_max
        self._accel_max = accel_max
        self._v_prev = 0

        self._vehicle_plot = None
        if control is not None:
            self.add_driver(control)
        self._animation = animation

        self._dt = dt
        self._t = 0
        self._stopsim = False

        self._verbose = verbose
        self._plot = False

        self._dim = dim

        self._x_hist = np.empty((0,len(x0)))

    def __str__(self):
        """
        String representation of vehicle (superclass method)

        :return: String representation of vehicle object
        :rtype: str
        """
        s = f"{self.__class__.__name__}: "
        s += f"x = {self._x}"
        return s

    @property
    def x(self):
        """
        Get vehicle state/configuration (superclass method)

        :return: Vehicle state :math:`(x, y, \theta)`
        :rtype: ndarray(3)
        """
        return self._x

    @property
    def x0(self):
        """
        Get vehicle initial state/configuration (superclass method)

        :return: Vehicle state :math:`(x, y, \theta)`
        :rtype: ndarray(3)

        The state is set to this value at the beginning of each simulation
        run.

        Set by ``Vehicle`` subclass constructor.

        :seealso: :func:`run`
        """
        return self._x0

    @property
    def x_hist(self):
        """
        Get vehicle state/configuration history (superclass method)

        :return: Vehicle state history
        :rtype: ndarray(n,3)

        The state at each time step resulting from a simulation
        run.

        :seealso: :func:`run`
        """
        return self._x_hist

    @property
    def speed_max(self):
        """
        Get maximum speed of vehicle (superclass method)

        :return: maximum speed
        :rtype: float

        Set by ``Vehicle`` subclass constructor.
        """
        return self._speed_max

    @property
    def accel_max(self):
        """
        Get maximum acceleration of vehicle (superclass method)

        :return: maximum acceleration
        :rtype: float

        Set by ``Vehicle`` subclass constructor.
        """
        return self._accel_max

    @property
    def dt(self):
        """
        Get sample time (superclass method)

        :return: discrete time step for simulation
        :rtype: float

        Set by ``Vehicle`` subclass constructor.

        :seealso: :func:`run`
        """
        return self._dt

    @property
    def verbose(self):
        """
        Get verbosity (superclass method)

        :return: verbosity level
        :rtype: bool

        Set by ``Vehicle`` subclass constructor.
        """
        return self._verbose

    @verbose.setter
    def verbose(self, verbose):
        """
        Set verbosity (superclass method)

        :return: verbosity level
        :rtype: bool

        Set by ``Vehicle`` subclass constructor.
        """
        self._verbose = verbose

    @property
    def control(self):
        """
        Get vehicle control (superclass method)

        :return: current control
        :rtype: 2-tuple, callable, interp1d or VehicleDriver

        Control can be:

            * a constant tuple as the control inputs to the vehicle
            * a function called as f(vehicle, t, x) that returns a tuple
            * an interpolator called as f(t) that returns a tuple, see
              SciPy interp1d
            * a driver agent, subclass of :func:`VehicleDriver`

        :seealso: :func:`eval_control`
        """
        return self._control

    @control.setter
    def control(self, control):
        """
        Set vehicle control (superclass method)

        :param control: new control
        :type control: 2-tuple, callable, interp1d or VehicleDriver

        Control can be:

            * a constant tuple as the control inputs to the vehicle
            * a function called as ``f(vehicle, t, x)`` that returns a tuple
            * an interpolator called as f(t) that returns a tuple, see
              SciPy interp1d
            * a driver agent, subclass of :func:`VehicleDriver`

        Example:

        .. runblock:: pycon

            >>> bike = Bicycle()
            >>> bike.control = RandomPath(10)
            >>> print(bike)

        :seealso: :func:`eval_control`, :func:`RandomPath`, :func:`PurePursuit`
        """
        self._control = control
        if isinstance(control, VehicleDriver):
            # if this is a driver agent, connect it to the vehicle
            control.vehicle = self


    # This function is overridden by the child class
    @abstractmethod
    def deriv(self, x, u):
        pass

    def add_driver(self, driver):
        """
        Add a driver agent (superclass method)

        :param driver: a driver agent object
        :type driver: VehicleDriver subclass

        .. warning: Deprecated.  Use ``vehicle.control = driver`` instead.

        :seealso: :func:`RandomPath`
        """

        warnings.warn('add_driver is deprecated, use veh.control=driver instead')
        self._control = driver
        driver._veh = self


    def run(self, N=1000, x0=None, control=None, animation=None, plot=True):
        """
        Simulate motion of vehicle (superclass method)

        :param N: Number of simulation steps, defaults to 1000
        :type N: int, optional
        :param x0: Initial state, defaults to value given to Vehicle constructor
        :type x0: array_like(3) or array_like(2)
        :param animation: vehicle animation object, defaults to None
        :type animation: VehicleAnimation subclass, optional
        :param plot: Enable plotting, defaults to False
        :type plot: bool, optional
        :return: State trajectory, each row is :math:`(x,y,\theta)`.
        :rtype: ndarray(n,3)

        Runs the vehicle simulation for ``N`` timesteps and optionally plots
        an animation.

        The control inputs are provied by ``control`` which can be:

            * a constant tuple as the control inputs to the vehicle
            * a function called as ``f(vehicle, t, x)`` that returns a tuple
            * an interpolator called as f(t) that returns a tuple, see
              SciPy interp1d
            * a driver agent, subclass of :func:`VehicleDriver`

        
        The simulation can be stopped prematurely by the control function
        calling :func:`stopif`.
        
        :seealso: :func:`init`, :func:`step`, :func:`control`
        """

        self.init(plot=plot, control=control, animation=animation, x0=x0)
            
        for i in range(N):
            self.step()

            # do the graphics
            if self._plot:
                if self._animation:
                    self._animation.update(self._x)
                if self._timer is not None:
                    self._timer.set_text(f"t = {self._t:.2f}")
                plt.pause(self._dt)

            # check for user requested stop
            if self._stopsim:
                print('USER REEQUESTED STOP AT time', self._t)
                break

        return self._x_hist

    def init(self, x0=None, animation=None, plot=False, control=None):
        """
        Initialize for simulation (superclass method)

        :param x0: Initial state, defaults to value given to Vehicle constructor
        :type x0: array_like(3) or array_like(2)
        :param animation: vehicle animation object, defaults to None
        :type animation: VehicleAnimation subclass, optional
        :param plot: Enable plotting, defaults to False
        :type plot: bool, optional

        Performs the following initializations:

            #. Clears the state history
            #. Sets state :math:`x = x_0`
            #. If a driver is attached, initialize it
            #. If plotting is enabled, initialize that

        If ``plot`` is set and no animation object is given, use a default
        ``VehiclePolygon('car')``

        :seealso: :func:`VehicleAnimation`
        """
        if x0 is not None:
            self._x = base.getvector(x0, 3)
        else:
            self._x = self._x0

        self._x_hist = np.empty((0,3))

        if control is not None:
            self._control = control

        self._t = 0

        # initialize the graphics
        self._plot = plot
        if plot:

            if animation is None:
                animation = self._animation  # get default animation if set
            else:
                # use default animation
                animation = VehiclePolygon("car")
            self._animation = animation

            # setu[ the plot]
            plt.clf()

            self._ax = base.plotvol2(self._dim)
        
            plt.xlabel('x')
            plt.ylabel('y')
            self._ax.set_aspect('equal')
            self._ax.figure.canvas.set_window_title(f"Robotics Toolbox for Python (Figure {self._ax.figure.number})")

            animation.add()  # add vehicle animation to axis
            self._timer = plt.figtext(0.85, 0.95, '')  # display time counter

        # initialize the driver
        if isinstance(self._control, VehicleDriver):
            self._control.init(ax=self._ax)

    def step(self, u1=None, u2=None):
        """
        Step simulator by one time step (superclass method)

        :return: odometry :math:`(\delta_d, \delta_\theta)`
        :rtype: ndarray(2)

        - ``veh.step(vel, steer)`` for a Bicycle vehicle model
        - ``veh.step((vel, steer))`` as above but control is a tuple
        - ``veh.step(vel, vel_diff)`` for a Unicycle vehicle model
        - ``veh.step()`` as above but control is taken from the ``control``
          attribute which might be a function or driver agent.

        #. Integrates the vehicle forward one timestep
        #. Updates the stored state and state history
        #. Returns the odometry

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import Bicycle
            >>> bike = Bicycle()  # default bicycle model
            >>> bike.step(1, 0.2)  # one step: v=1, γ=0.2
            >>> bike.x
            >>> bike.step(1, 0.2)  # another step: v=1, γ=0.2
            >>> bike.x

        .. note:: Vehicle control input limits are applied.

        :seealso: :func:`control`, :func:`update`, :func:`run`
        """
        # determine vehicle control
        if u1 is not None:
            if u2 is not None:
                u = self.eval_control((u1, u2), self._x)
            else:
                u = self.eval_control(u1, self._x)
        else:
            u = self.eval_control(self._control, self._x)

        # update state (used to be function control() in MATLAB version)
        xd = self._dt * self.deriv(self._x, u)  # delta state

        # update state vector
        self._x += xd
        self._x_hist = np.vstack((self._x_hist, self._x))

        # odometry comes from change in state vector
        odo = np.r_[np.linalg.norm(xd[0:2]), xd[2]]

        if self._V is not None:
            odo += linalg.sqrtm(self._V) @ np.random.randn(2)

        self._t += self._dt

        # be verbose
        if self._verbose:
            print(f"{self._t:8.2f}: u=({u[0]:8.2f}, {u[1]:8.2f}), x=({self._x[0]:8.2f}, {self._x[1]:8.2f}, {self._x[2]:8.2f})")

        return odo


    def eval_control(self, control, x):
        """
        Evaluate vehicle control input (superclass method)

        :param control: vehicle control
        :type control: [type]
        :param x: vehicle state :math:`(x, y, \theta)`
        :type x: array_like(3)
        :raises ValueError: bad control
        :return: vehicle control inputs
        :rtype: ndarray(2)

        Evaluates the control for this time step and state. Control can be:

            * a constant 2-tuple as the control inputs to the vehicle
            * a function called as ``f(vehicle, t, x)`` that returns a tuple
            * an interpolator called as f(t) that returns a tuple, see
              SciPy interp1d
            * a ``VehicleDriver`` subclass object

        .. note:: Vehicle steering, speed and acceleration limits are applied.
        """
        # was called control() in the MATLAB version

        if base.isvector(control, 2):
            # control is a constant
            u = base.getvector(control, 2)
        
        elif isinstance(control, VehicleDriver):
            # vehicle has a driver object
            u = control.demand()

        elif isinstance(control, interpolate.interpolate.interp1d):
            # control is an interp1d object
            u = control(self._t)

        elif callable(control):
            # control is a user function of time and state
            u = control(self, self._t, x)

        else:
            raise ValueError('bad control specified')

        # apply limits
        ulim = self.u_limited(u)
        return ulim

    def stopif(self, stop):
        """
        Stop the simulation (superclass method)

        :param stop: stop condition
        :type stop: bool

        A control function can stop the simulation initated by the ``run``
        method if ``stop`` is True.

        :seealso: :func:`run`
        """
        if stop:
            self._stopsim = True

    def plot(self, path=None, block=True):
        """
        [summary] (superclass method)

        :param path: [description], defaults to None
        :type path: [type], optional
        :param block: [description], defaults to True
        :type block: bool, optional
        """
 
        plt.plot(path[:,0], path[:,1])
        plt.show(block=block)

    def plot_x_y(self, block=True, **kwargs):
        xyt = self._x_hist
        plt.plot(xyt[:,0], xyt[:, 1], **kwargs)
        plt.show(block=block)

    def plot_xyt_t(self, block=True, **kwargs):
        xyt = self._x_hist
        t = np.arange(0, xyt.shape[0] * self._dt, self._dt)
        plt.plot(xyt[:,0], xyt[:, :], **kwargs)
        plt.legend(['x', 'y', '$\\theta$'])
        plt.show(block=block)

    def limits_va(self, v):
        """
        Apply velocity and acceleration limits (superclass method)

        :param v: commanded velocity
        :type v: float
        :return: allowed velocity
        :rtype: float

        .. note:: This function is stateful, requires previous velocity,
            ``_v_prev`` attribute, to enable acceleration limiting.  This
            is reset at the start of each simulation.
        """
        # acceleration limit
        if (v - self._v_prev) / self._dt > self._accel_max:
            v = self._v_prev + self._accelmax * self._dt;
        elif (v - self._v_prev) / self._dt < -self._accel_max:
            v = self._v_prev - self._accel_max * self._dt;
        self._v_prev = v
        
        # speed limit
        return min(self._speed_max, max(v, -self._speed_max));


    def path(self, t=10, u=None, x0=None):
        """
        Compute path by integration (superclass method)

        :param t: [description], defaults to None
        :type t: [type], optional
        :param u: [description], defaults to None
        :type u: [type], optional
        :param x0: initial state, defaults to (0,0,0)
        :type x0: array_like(3), optional
        :return: time vector and state history
        :rtype: (ndarray(1), ndarray(n,3))

                    % XF = V.path(TF, U) is the final state of the vehicle (3x1) from the initial
            % state (0,0,0) with the control inputs U (vehicle specific).  TF is  a scalar to 
            % specify the total integration time.
            %
            % XP = V.path(TV, U) is the trajectory of the vehicle (Nx3) from the initial
            % state (0,0,0) with the control inputs U (vehicle specific).  T is a vector (N) of 
            % times for which elements of the trajectory will be computed.
            %
            % XP = V.path(T, U, X0) as above but specify the initial state.
            %
            % Notes::
            % - Integration is performed using ODE45.
            % - The ODE being integrated is given by the deriv method of the vehicle object.

             # t, x = veh.path(5, u=control)
    # print(t)
        """
        if x0 is None:
            x0 = np.zeros(3)

        def dynamics(t, x, vehicle, demand):
            u = vehicle.control(demand, x)
            
            return vehicle.deriv(x, u)

        if base.isscalar(t):
            t_span = (0, t)
            t_eval = np.linspace(0, t, 100)
        elif isinstance(t, np.ndarray):
            t_span = (t[0], t[-1])
            t_eval = t
        else:
            raise ValueError('bad time argument')
        sol = integrate.solve_ivp(dynamics, t_span, x0, t_eval=t_eval, method="RK45", args=(self, u))

        return (sol.t, sol.y)
# ========================================================================= #

class Bicycle(Vehicle):

    def __init__(self,
                l=1,
                steer_max=0.45 * pi,
                **kwargs
                ):
        r"""
        Create new bicycle kinematic model

        :param l: wheel base, defaults to 1
        :type l: float, optional
        :param steer_max: [description], defaults to :math:`0.45\pi`
        :type steer_max: float, optional
        :param **kwargs: additional arguments passed to :class:`Vehicle`
            constructor

        :seealso: :class:`.Vehicle`
        """
        super().__init__(**kwargs)

        self._l = l
        self._steer_max = steer_max

    def __str__(self):

        s = super().__str__()
        s += f"\n  L={self._l}, steer_max={self._steer_max}, speed_max={self._speed_max}, accel_max={self._accel_max}"
        return s

    @property
    def l(self):
        """
        Vehicle wheelbase

        :return: vehicle wheelbase
        :rtype: float
        """
        return self._l

    @property
    def steer_max(self):
        """
        Vehicle maximum steering wheel angle

        :return: maximum angle
        :rtype: float
        """
        return self._steer_max


    def f(self, x, odo, w=None):
        r"""
        Predict next state based on odometry

        :param x: vehicle state :math:`(x, y, \theta)`
        :type x: array_like(3)
        :param odo: vehicle odometry :math:`(\delta_d, \delta_\theta)`
        :type odo: array_like(2)
        :param w: [description], defaults to (0,0)
        :type w: array_like(2), optional
        :return: predicted vehicle state
        :rtype: ndarray(3)

        Returns the predicted next state based on current state and odometry 
        value.  ``w`` is a random variable that represents additive
        odometry noise for simulation purposes.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import Bicycle
            >>> bike = Bicycle()  # default bicycle model
            >>> bike.f([0,0,0], [0.2, 0.1])

        .. note:: This is the state update equation used for EKF localization.
        """
        x = base.getvector(x, 3)
        odo = base.getvector(odo, 2)

        dd = odo[0]
        dth = odo[1]
        thp = x[2]

        if w is not None:
            w = base.getvector(w, 2)
            dd += w[0]
            dth += w[1]

        # TODO not sure when vectorized version is needed
        return x + np.r_[dd * np.cos(thp), dd * np.sin(thp), dth]

    def Fx(self, x, odo):
        r"""
        Jacobian df/dx

        :param x: vehicle state :math:`(x, y, \theta)`
        :type x: array_like(3)
        :param odo: vehicle odometry :math:`(\delta_d, \delta_\theta)`
        :type odo: array_like(2)
        :return: Jacobian matrix 
        :rtype: ndarray(3,3)
        
        Returns the Jacobian matrix :math:`\frac{\partial \vec{f}}{\partial \vec{x}}` for
        the given odometry.

        :seealso: :func:`Bicycle.f`, :func:`Bicycle.Fv`
        """
        dd = odo[0]
        dth = odo[1]
        thp = x[2] + dth

        J = np.array([
                [1,   0,  -dd * sin(thp)],
                [0,   1,   dd * cos(thp)],
                [0,   0,   1],
            ])
        return J

    def Fv(self, x, odo):
        r"""
        Jacobian df/dv

        :param x: vehicle state :math:`(x, y, \theta)`
        :type x: array_like(3)
        :param odo: vehicle odometry :math:`(\delta_d, \delta_\theta)`
        :type odo: array_like(2)
        :return: Jacobian matrix 
        :rtype: ndarray(3,2)
        
        Returns the Jacobian matrix :math:`\frac{\partial \vec{f}}{\partial \vec{v}}` for
        the given odometry.

        :seealso: :func:`Bicycle.f`, :func:`Bicycle.Fx`
        """

        dd = odo[0]
        dth = odo[1]
        thp = x[2]


        J = np.array([
                [cos(thp),    0],
                [sin(thp),    0],
                [0,           1],
            ])

    def deriv(self, x, u):
        r"""
        Time derivative of state

        :param x: vehicle state :math:`(x, y, \theta)`
        :type x: array_like(3)
        :param u: control input
        :type u: array_like(2)
        :return: state derivative :math:`(\dot{x}, \dot{y}, \dot{\theta})`
        :rtype: ndarray(3)

        Returns the time derivative of state (3x1) at the state ``x`` with 
        inputs ``u``.

        .. note:: Vehicle speed and steering limits are not applied here
        """
        
        # unpack some variables
        theta = x[2]
        v = u[0]
        gamma = u[1]
            
        return v * np.r_[
                cos(theta), 
                sin(theta), 
                tan(gamma) / self.l
                    ]

    def u_limited(self, u):

        # limit speed and steer angle
        ulim = np.array(u)
        ulim[0] = self.limits_va(u[0])
        ulim[1] = np.maximum(-self._steer_max, np.minimum(self._steer_max, u[1]))

        return ulim

# ========================================================================= #

class Unicycle(Vehicle):

    def __init__(self,
                w=1,
                **kwargs):
        r"""
        Create new unicycle kinematic model

        :param w: vehicle width, defaults to 1
        :type w: float, optional
        :param **kwargs: additional arguments passed to :class:`Vehicle`
            constructor

        :seealso: :class:`.Vehicle`
        """
        super().__init__(**kwargs)
        self._w = w

    def __str__(self):

        s = super().__str__()
        s += f"\n  W={self._w}, steer_max={self._steer_max}, vel_max={self._vel_max}, accel_max={self.accel_max}"
        return s

    def f(self, x=None, odo=None, w=None):
        
        if w is None:
            w = np.array([0, 0])

        dd = odo[0] + w[0]
        dth = odo[1] + w[1]
        thp = x[:, 2]
        x_next = x + [dd * np.cos(thp), dd * np.sin(thp), np.ones(np.size(x, 0)*dth)]

        return x_next

    def Fx(self, x, odo):
        r"""
        Jacobian df/dx

        :param x: vehicle state :math:`(x, y, \theta)`
        :type x: array_like(3)
        :param odo: vehicle odometry :math:`(\delta_d, \delta_\theta)`
        :type odo: array_like(2)
        :return: Jacobian matrix 
        :rtype: ndarray(3,3)
        
        Returns the Jacobian matrix :math:`\frac{\partial \vec{f}}{\partial \vec{x}}` for
        the given odometry.

        :seealso: :func:`Bicycle.f`, :func:`Bicycle.Fv`
        """
        dd = odo[0]
        dth = odo[1]
        thp = x[2] + dth

        J = np.array([
                [1,   0,  -dd * sin(thp)],
                [0,   1,   dd * cos(thp)],
                [0,   0,   1],
            ])
        return J

    def Fv(self, x, odo):
        r"""
        Jacobian df/dv

        :param x: vehicle state :math:`(x, y, \theta)`
        :type x: array_like(3)
        :param odo: vehicle odometry :math:`(\delta_d, \delta_\theta)`
        :type odo: array_like(2)
        :return: Jacobian matrix 
        :rtype: ndarray(3,2)
        
        Returns the Jacobian matrix :math:`\frac{\partial \vec{f}}{\partial \vec{v}}` for
        the given odometry.

        :seealso: :func:`Bicycle.f`, :func:`Bicycle.Fx`
        """

        dd = odo[0]
        dth = odo[1]
        thp = x[2]


        J = np.array([
                [cos(thp),    0],
                [sin(thp),    0],
                [0,           1],
            ])

    def deriv(self, t, x, u):
        r"""
        Time derivative of state

        :param x: vehicle state :math:`(x, y, \theta)`
        :type x: array_like(3)
        :param u: control input
        :type u: array_like(2)
        :return: state derivative :math:`(\dot{x}, \dot{y}, \dot{\theta})`
        :rtype: ndarray(3)

        Returns the time derivative of state (3x1) at the state ``x`` with 
        inputs ``u``.

        .. note:: Vehicle speed and steering limits are not applied here
        """
        
        # unpack some variables
        theta = x[2]
        v = u[0]
        vdiff = u[1]

        return np.r_[
                v * cos(theta), 
                v * sin(theta), 
                vdiff / self.w
                    ]

    def u_limited(self, u):

        # limit speed and steer angle
        ulim = np.array(u)
        ulim[0] = self.limits_va(u[0])
        ulim[1] = np.maximum(-self._steer_max, np.minimum(self._steer_max, u[1]))

        return ulim


if __name__ == "__main__":
    from math import pi

    V = np.diag(np.r_[0.02, 0.5 * pi / 180] ** 2)

    v = VehiclePolygon()
    # v = VehicleIcon('greycar2', scale=2, rotation=90)

    veh = Bicycle(covar=V, animation=v, control=RandomPath(10), verbose=False)
    print(veh)

    odo = veh.step(1, 0.3)
    print(odo)

    print(veh.x)

    print(veh.f([0, 0, 0], odo))

    def control(v, t, x):
        goal = (6,6)
        goal_heading = atan2(goal[1]-x[1], goal[0]-x[0])
        d_heading = base.angdiff(goal_heading, x[2])
        v.stopif(base.norm(x[0:2] - goal) < 0.1)

        return (1, d_heading)

    veh.control=RandomPath(10)
    p = veh.run(1000, plot=True)
    # plt.show()
    print(p)

    veh.plot_xyt_t()
    # veh.plot(p)

    # t, x = veh.path(5, u=control)
    # print(t)

    # fig, ax = plt.subplots()

    # ax.set_xlim(-5, 5)
    # ax.set_ylim(-5, 5)


    # v = VehicleAnimation.Polygon(shape='triangle', maxdim=0.1, color='r')
    # v = VehicleAnimation.Icon('car3.png', maxdim=2, centre=[0.3, 0.5])
    # v = VehicleAnimation.Icon('/Users/corkep/Dropbox/code/robotics-toolbox-python/roboticstoolbox/data/car1.png', maxdim=2, centre=[0.3, 0.5])
    # v = VehicleAnimation.icon('car3.png', maxdim=2, centre=[0.3, 0.5])
    # v = VehicleAnimation.marker()
    # v.start()
    # plt.grid(True)
    # # plt.axis('equal')

    # for theta in np.linspace(0, 2 * np.pi, 100):
    #     v.update([0, 0, theta])
    #     plt.pause(0.1)


import numpy as np
from spatialmath import SE3
import roboticstoolbox as rtb
import timeit
from ansitable import ANSITable, Column
import traceback

# change for the robot IK under test, must set:
#  * robot, the DHRobot object
#  * T, the end-effector pose
#  * q0, the initial joint angles for solution

example = 'puma'  # 'panda'

if example == 'puma':
    # Puma robot case
    robot = rtb.models.DH.Puma560()
    q = robot.qn
    q0 = robot.qz
    T = robot.fkine(q)
elif example == 'panda':
    # Panda robot case
    robot = rtb.models.DH.Panda()
    T = SE3(0.7, 0.2, 0.1) * SE3.OA([0, 1, 0], [0, 0, -1])
    q0 = robot.qz

solvers = [
        'Nelder-Mead',
        'Powell',
        'CG',
        'BFGS',
        'Newton-CG',  ## Jacobian is required
        'L-BFGS-B',
        'TNC',
        'COBYLA',
        'SLSQP',
        'trust-constr',
        'dogleg',
        'trust-ncg',
        'trust-exact',
        'trust-krylov',
    ]


# setup to run timeit
setup = '''
from __main__ import robot, T, q0
'''
N = 10

# setup results table
table = ANSITable(
    Column("Solver", headalign="^", colalign='<'),
    Column("Time (ms)", headalign="^", fmt="{:.2g}", colalign='>'),
    Column("Error", headalign="^", fmt="{:.3g}", colalign='>'),
    border="thick")

# test the IK methods
for solver in solvers:
    print('Testing:', solver)
    
    # test the method, don't pass q0 to the analytic function
    try:
        sol = robot.ikine_min(T, q0=q0, qlim=True, method=solver)
    except Exception as e:
        print('***', solver, ' failed')
        print(e)
        continue

    # print error message if there is one
    if not sol.success:
        print('  failed:', sol.reason)

    # evalute the error
    err = np.linalg.norm(T - robot.fkine(sol.q))
    print('  error', err)

    if N > 0:
        # evaluate the execution time
        t = timeit.timeit(stmt=f"robot.ikine_min(T, q0=q0, qlim=True, method='{solver}')", setup=setup, number=N)
    else:
        t = 0

    # add it to the output table
    table.row(f"`{solver}`", t/N*1e3, err)

# pretty print the results     
table.print()
print(table.markdown())



#!/usr/bin/env python
"""
@author Jesse Haviland
"""
import time
import roboticstoolbox as rp
import numpy as np
from roboticstoolbox.backends.Connector import Connector

from roboticstoolbox.backends.PyPlot.RobotPlot import RobotPlot
from roboticstoolbox.backends.PyPlot.EllipsePlot import EllipsePlot
from spatialmath.base.argcheck import getvector
# from roboticstoolbox.tools import Ticker

_mpl = False

try:
    import matplotlib
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    from matplotlib.widgets import Slider
    matplotlib.rcParams['pdf.fonttype'] = 42
    matplotlib.rcParams['ps.fonttype'] = 42
    plt.style.use('ggplot')
    matplotlib.rcParams['font.size'] = 7
    matplotlib.rcParams['lines.linewidth'] = 0.5
    matplotlib.rcParams['xtick.major.size'] = 1.5
    matplotlib.rcParams['ytick.major.size'] = 1.5
    matplotlib.rcParams['axes.labelpad'] = 1
    plt.rc('grid', linestyle="-", color='#dbdbdb')
    _mpl = True
except ImportError:    # pragma nocover
    pass


class PyPlot(Connector):
    """
    Graphical backend using matplotlib

    matplotlib is a common and highly portable graphics library for Python,
    but has relatively limited 3D capability.

    Example:

    .. code-block:: python
        :linenos:

        import roboticstoolbox as rtb

        robot = rtb.models.DH.Panda()  # create a robot

        pyplot = rtb.backends.PyPlot()  # create a PyPlot backend
        pyplot.add(robot)              # add the robot to the backend
        robot.q = robot.qz             # set the robot configuration
        pyplot.step()                  # update the backend and graphical view

    .. note::  PyPlot is the default backend, and ``robot.plot(q)`` effectively
        performs lines 7-8 above.

    """

    def __init__(self):

        super(PyPlot, self).__init__()
        self.robots = []
        self.ellipses = []

        if not _mpl:    # pragma nocover
            raise ImportError(
                '\n\nYou do not have matplotlib installed, do:\n'
                'pip install matplotlib\n\n')

    def launch(self, name=None, limits=None):
        """
        Launch a graphical interface

        ```env = launch()``` creates a blank 3D matplotlib figure and returns
        a reference to the backend.
        """

        super().launch()

        self.limits = limits
        if limits is not None:
            self.limits = getvector(limits, 6)

        projection = 'ortho'
        labels = ['X', 'Y', 'Z']

        if name is not None:
            self.fig = plt.figure(name)
        else:
            self.fig = plt.figure('Robotics Toolbox for Python')

        self.fig.subplots_adjust(left=-0.09, bottom=0, top=1, right=0.99)

        # Create a 3D axes
        self.ax = self.fig.add_subplot(
            111, projection='3d', proj_type=projection)
        self.ax.set_facecolor('white')

        self.ax.set_xbound(-0.5, 0.5)
        self.ax.set_ybound(-0.5, 0.5)
        self.ax.set_zbound(0.0, 0.5)

        self.ax.set_xlabel(labels[0])
        self.ax.set_ylabel(labels[1])
        self.ax.set_zlabel(labels[2])

        if limits is not None:
            self.ax.set_xlim3d([limits[0], limits[1]])
            self.ax.set_ylim3d([limits[2], limits[3]])
            self.ax.set_zlim3d([limits[4], limits[5]])

        # disable the display of value under cursor
        self.ax.format_coord = lambda x, y: ''

        # add time display in top-right corner
        self.timer = plt.figtext(0.85, 0.95, '')

        if _isnotebook():
            plt.ion()
            self.fig.canvas.draw()
        else:
            plt.ion()
            plt.show()

        self.sim_time = 0

        # # Set the signal handler and a 0.1 second plot updater
        # signal.signal(signal.SIGALRM, self._plot_handler)
        # signal.setitimer(signal.ITIMER_REAL, 0.1, 0.1)
        # TODO still need to finish this, and get Jupyter animation working

    def step(self, dt=0.05):
        """
        Update the graphical scene

        :param dt: time step in seconds, defaults to 50 (0.05 s)
        :type dt: int, optional

        ``env.step(args)`` triggers an update of the 3D scene in the matplotlib
        window referenced by ``env``.

        .. note::

            - Each robot in the scene is updated based on
              their control type (position, velocity, acceleration, or torque).
            - Upon acting, the other three of the four control types will be
              updated in the internal state of the robot object.
            - The control type is defined by the robot object, and not all
              robot objects support all control types.
            - Execution is blocked for the specified interval

        """

        super().step()

        self._step_robots(dt)

        # plt.ioff()

        self._draw_ellipses()
        self._draw_robots()
        self._set_axes_equal()

        # update time and display it on plot
        if self.sim_time > 0:
            self.timer.set_text(f"t = {self.sim_time:.2f}")
        self.sim_time += dt

        # plt.ion()

        if _isnotebook():
            plt.draw()
            self.fig.canvas.draw()
            time.sleep(dt)
        else:
            plt.draw()
            plt.pause(dt)

        self._update_robots()

    def reset(self):
        """
        Reset the graphical scene

        ``env.reset()`` triggers a reset of the 3D scene in the matplotlib
        window referenced by ``env``. It is restored to the original state
        defined by ``launch()``.
        """
        # TODO what does this actually do for matplotlib??

        super().reset()

    def restart(self):
        """
        Restart the graphics display

        ``env.restart()`` triggers a restart of the matplotlib view referenced
        by ``env``. It is closed and relaunched to the original state defined
        by ``launch()``.

        """
        # TODO what does this actually do for matplotlib??

        super().restart()

    def close(self):
        """
        ``env.close()`` gracefully closes the matplotlib window
        referenced by ``env``.
        """
        # TODO what does this actually do for matplotlib??

        super().close()

        # signal.setitimer(signal.ITIMER_REAL, 0)
        plt.close(self.fig)

    #
    #  Methods to interface with the robots created in other environments
    #

    def add(
            self, ob, readonly=False, display=True,
            jointaxes=True, eeframe=True, shadow=True, name=True):
        """
        Add a robot to the graphical scene

        :param ob: The object to add to the plot (robot or ellipse)
        :type ob: DHRobot or EllipsePlot
        :param readonly: Do not update the state of the object
            (i.e. display not simulate), defaults to False
        :type readonly: bool, optional
        :param display: Display the object, defaults to True
        :type display: bool, optional
        :param jointaxes: Show the joint axes of the robot with arrows,
            defaults to True
        :type jointaxes: bool, optional
        :param eeframe: Show the end-effector frame of the robot,
            defaults to True
        :type eeframe: bool, optional
        :param shadow: Display a shadow of the robot on the x-y gound plane,
            defaults to True
        :type shadow: bool, optional
        :param name: Display the name of the robot, defaults to True
        :type name: bool, optional

        ``id = env.add(robot)`` adds the ``robot`` to the graphical
            environment.

        .. note::

            - ``robot`` must be of an appropriate class.
            - Adds the robot object to a list of robots which will be updated
              when the ``step()`` method is called.

        """

        super().add()

        if isinstance(ob, rp.DHRobot) or isinstance(ob, rp.ERobot):
            self.robots.append(
                RobotPlot(
                    ob, self.ax, readonly, display,
                    jointaxes, eeframe, shadow, name))
            self.robots[len(self.robots) - 1].draw()
            id = len(self.robots)

        elif isinstance(ob, EllipsePlot):
            ob.ax = self.ax
            self.ellipses.append(ob)
            self.ellipses[len(self.ellipses) - 1].draw()
            id = len(self.ellipses)

        plt.draw()
        plt.show(block=False)

        self._set_axes_equal()
        return id

    def remove(self):
        """
        Remove a robot to the graphical scene

        :param id: The id of the robot to remove. Can be either the DHLink or
            GraphicalRobot
        :type id: class:`~roboticstoolbox.robot.DHRobot.DHRobot`,
                  class:`roboticstoolbox.backends.VPython.graphics_robot.GraphicalRobot`
        :param fig_num: The canvas index to delete the robot from, defaults to
             the initial one
        :type fig_num: int, optional
        :raises ValueError: Figure number must be between 0 and total number
            of canvases
        :raises TypeError: Input must be a DHLink or GraphicalRobot

        ``env.remove(robot)`` removes the ``robot`` from the graphical
            environment.
        """
        # TODO should be an id to remove?

        super().remove()

    def hold(self):           # pragma: no cover
        '''
        hold() keeps the plot open i.e. stops the plot from closing once
        the main script has finished.

        '''

        # signal.setitimer(signal.ITIMER_REAL, 0)
        plt.ioff()

        try:
            plt.show()
        except AttributeError:
            pass

    #
    #  Private methods
    #

    def _step_robots(self, dt):

        for rpl in self.robots:
            robot = rpl.robot

            if rpl.readonly or robot.control_type == 'p':
                pass            # pragma: no cover

            elif robot.control_type == 'v':

                for i in range(robot.n):
                    robot.q[i] += robot.qd[i] * (dt)

            elif robot.control_type == 'a':     # pragma: no cover
                pass

            else:            # pragma: no cover
                # Should be impossible to reach
                raise ValueError(
                    'Invalid robot.control_type. '
                    'Must be one of \'p\', \'v\', or \'a\'')

    def _update_robots(self):
        pass

    def _draw_robots(self):

        for i in range(len(self.robots)):
            self.robots[i].draw()

    def _draw_ellipses(self):

        for i in range(len(self.ellipses)):
            self.ellipses[i].draw()

    # def _plot_handler(self, sig, frame):
    #     try:
    #         plt.pause(0.001)
    #     except(AttributeError):
    #         pass

    def _set_axes_equal(self):
        """
        Make axes of 3D plot have equal scale so that spheres appear as
        spheres, cubes as cubes, etc..  This is one possible solution to
        Matplotlib's ax.set_aspect('equal') and ax.axis('equal') not
        working for 3D.

        """

        if self.limits is not None:
            return

        self.ax.autoscale(enable=True, axis='both', tight=False)

        x_limits = self.ax.get_xlim3d()
        y_limits = self.ax.get_ylim3d()
        z_limits = self.ax.get_zlim3d()

        x_range = abs(x_limits[1] - x_limits[0])
        x_middle = np.mean(x_limits)
        y_range = abs(y_limits[1] - y_limits[0])
        y_middle = np.mean(y_limits)
        z_range = abs(z_limits[1] - z_limits[0])
        z_middle = np.mean(z_limits)

        # The plot bounding box is a sphere in the sense of the infinity
        # norm, hence I call half the max range the plot radius.
        plot_radius = 0.5*max([x_range, y_range, z_range])

        self.ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
        self.ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
        self.ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])

    def _add_teach_panel(self, robot):

        if _isnotebook():
            raise RuntimeError('cannot use teach panel under Jupyter')

        fig = self.fig

        # Add text to the plots
        def text_trans(text):  # pragma: no cover
            T = robot.fkine()
            t = np.round(T.t, 3)
            r = np.round(T.rpy('deg'), 3)
            text[0].set_text("x: {0}".format(t[0]))
            text[1].set_text("y: {0}".format(t[1]))
            text[2].set_text("z: {0}".format(t[2]))
            text[3].set_text("r: {0}".format(r[0]))
            text[4].set_text("p: {0}".format(r[1]))
            text[5].set_text("y: {0}".format(r[2]))

        # Update the self state in mpl and the text
        def update(val, text, robot):  # pragma: no cover
            for i in range(robot.n):
                robot.q[i] = self.sjoint[i].val * np.pi/180

            text_trans(text)

            # Step the environment
            self.step(0)

        fig.subplots_adjust(left=0.25)
        text = []

        x1 = 0.04
        x2 = 0.22
        yh = 0.04
        ym = 0.5 - (robot.n * yh) / 2 + 0.17/2

        self.axjoint = []
        self.sjoint = []

        qlim = np.copy(robot.qlim) * 180/np.pi

        if np.all(qlim == 0):     # pragma: no cover
            qlim[0, :] = -180
            qlim[1, :] = 180

        # Set the pose text
        T = robot.fkine()
        t = np.round(T.t, 3)
        r = np.round(T.rpy(), 3)

        fig.text(
            0.02,  1 - ym + 0.25, "End-effector Pose",
            fontsize=9, weight="bold", color="#4f4f4f")
        text.append(fig.text(
            0.03, 1 - ym + 0.20, "x: {0}".format(t[0]),
            fontsize=9, color="#2b2b2b"))
        text.append(fig.text(
            0.03, 1 - ym + 0.16, "y: {0}".format(t[1]),
            fontsize=9, color="#2b2b2b"))
        text.append(fig.text(
            0.03, 1 - ym + 0.12, "z: {0}".format(t[2]),
            fontsize=9, color="#2b2b2b"))
        text.append(fig.text(
            0.15, 1 - ym + 0.20, "r: {0}".format(r[0]),
            fontsize=9, color="#2b2b2b"))
        text.append(fig.text(
            0.15, 1 - ym + 0.16, "p: {0}".format(r[1]),
            fontsize=9, color="#2b2b2b"))
        text.append(fig.text(
            0.15, 1 - ym + 0.12, "y: {0}".format(r[2]),
            fontsize=9, color="#2b2b2b"))
        fig.text(
            0.02,  1 - ym + 0.06, "Joint angles",
            fontsize=9, weight="bold", color="#4f4f4f")

        for i in range(robot.n):
            ymin = (1 - ym) - i * yh
            self.axjoint.append(
                fig.add_axes([x1, ymin, x2, 0.03], facecolor='#dbdbdb'))

            self.sjoint.append(
                Slider(
                    self.axjoint[i], 'q' + str(i),
                    qlim[0, i], qlim[1, i], robot.q[i] * 180/np.pi))

            self.sjoint[i].on_changed(lambda x: update(x, text, robot))


def _isnotebook():
    """
    Determine if code is being run from a Jupyter notebook

    ``_isnotebook`` is True if running Jupyter notebook, else False

    :references:

        - https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-
        is-executed-in-the-ipython-notebook/39662359#39662359
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter



#!/usr/bin/env python
"""
@author Micah Huth
"""

from vpython import vec, vertex, color, triangle, compound
from roboticstoolbox.backends.VPython.common_functions import \
    vector, x_axis_vector
from io import StringIO

try:
    from roboticstoolbox.tools.stdout_supress import pipes
except Exception:  # pragma nocover
    from contextlib import contextmanager

    @contextmanager
    def pipes(stdout=None, stderr=None):
        pass

_out = StringIO()
_err = StringIO()
try:
    with pipes(stdout=_out, stderr=_err):
        from stl import mesh
except Exception:  # pragma nocover
    from stl import mesh


def import_object_from_numpy_stl(filename, scene):  # pragma nocover
    """
    Import either an ASCII or BINARY file format of an STL file.
    The triangles will be combined into a single compound entity.

    :param filename: Path of the stl file to import.
    :type filename: `str`
    :param scene: The scene in which to draw the object
    :type scene: class:`vpython.canvas`
    :return: Compound object of a collection of triangles formed from an
        stl file.
    :rtype: class:`vpython.compound`
    """
    # filename = meshdata[0]
    # scale = meshdata[1]
    # origin = meshdata[2]
    # origin_pos = get_pose_pos(origin)
    # origin_x = get_pose_x_vec(origin)
    # origin_y = get_pose_y_vec(origin)
    # origin_z = get_pose_z_vec(origin)

    # Load the mesh using NumPy-STL
    the_mesh = mesh.Mesh.from_file(filename)

    num_faces = len(the_mesh.vectors)
    triangles = []

    # For every face in the model
    for face in range(0, num_faces):
        # Get the (3) 3D points
        point0 = the_mesh.vectors[face][0]
        point1 = the_mesh.vectors[face][1]
        point2 = the_mesh.vectors[face][2]

        # Get the normal direction for the face
        normal0 = the_mesh.normals[face][0]
        normal1 = the_mesh.normals[face][1]
        normal2 = the_mesh.normals[face][2]
        normal = vec(normal0, normal1, normal2)

        # Create the VPython 3D points
        vertex0 = vertex(
            pos=vec(point0[0], point0[1], point0[2]),
            normal=normal,
            color=color.white
        )
        vertex1 = vertex(
            pos=vec(point1[0], point1[1], point1[2]),
            normal=normal,
            color=color.white
        )
        vertex2 = vertex(
            pos=vec(point2[0], point2[1], point2[2]),
            normal=normal,
            color=color.white
        )

        # Combine them in a list
        vertices = [vertex0, vertex1, vertex2]

        # Create a triangle using the points, and add it to the list
        triangles.append(triangle(canvas=scene, vs=vertices))

    # Return a compound of the triangles
    visual_mesh = compound(triangles, origin=vec(0, 0, 0), canvas=scene)
    # visual_mesh.length *= scale[0]
    # visual_mesh.height *= scale[1]
    # visual_mesh.width *= scale[2]

    return visual_mesh


def set_stl_origin(
        stl_obj, current_obj_origin,
        required_obj_origin, scene):  # pragma nocover
    """
    Move the object so the required origin is at (0, 0, 0). Then set the
    origin for the generated stl object. Origin can't be changed, so creating
    a compound of itself allows setting an origin location

    :param stl_obj: The generated stl object.
    :type stl_obj: class:`vpython.compound`
    :param current_obj_origin: Current coordinates of the origin of the model
    :type current_obj_origin: class:`vpython.vector`
    :param required_obj_origin: Required coordinates to place the origin
        at (0, 0, 0)
    :type required_obj_origin: class:`vpython.vector`
    :param scene: The scene in which to draw the object
    :type scene: class:`vpython.canvas`
    :return: Compound object of itself, with origin set respective to the joint
    :rtype: class:`vpython.compound`
    """
    # Move the object to put the origin at 0, 0, 0
    movement = required_obj_origin - current_obj_origin
    stl_obj.pos += movement

    # Set invisible to return an overwritten copy
    stl_obj.visible = False

    # Return a compound of itself with the origin at (0, 0, 0)
    return compound(
        [stl_obj], origin=vector(0, 0, 0), vector=x_axis_vector, canvas=scene)



# Copyright (c) 2015, Open Source Robotics Foundation, Inc.
# Copyright (c) 2013, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Open Source Robotics Foundation, Inc.
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior
#       written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Authors: Stuart Glaser, William Woodall, Robert Haschke
# Maintainer: Morgan Quigley <morgan@osrfoundation.org>

import textwrap
from optparse import OptionParser, IndentedHelpFormatter
from .color import colorize, message


class ColoredOptionParser(OptionParser):   # pragma: no cover
    def error(self, message):
        msg = colorize(message, 'red')
        OptionParser.error(self, msg)


_original_wrap = textwrap.wrap


def wrap_with_newlines(text, width, **kwargs):  # pragma: no cover
    result = []
    for paragraph in text.split('\n'):
        result.extend(_original_wrap(paragraph, width, **kwargs))
    return result


class IndentedHelpFormatterWithNL(IndentedHelpFormatter):  # pragma: no cover
    def __init__(self, *args, **kwargs):
        IndentedHelpFormatter.__init__(self, *args, **kwargs)

    def format_option(self, text):
        textwrap.wrap, old = wrap_with_newlines, textwrap.wrap
        result = IndentedHelpFormatter.format_option(self, text)
        textwrap.wrap = old
        return result


def process_args(argv, require_input=True):   # pragma: no cover
    parser = ColoredOptionParser(usage="usage: %prog [options] <input>",
                                 formatter=IndentedHelpFormatterWithNL())
    parser.add_option("-o", dest="output", metavar="FILE",
                      help="write output to FILE instead of stdout")
    parser.add_option("--deps", action="store_true", dest="just_deps",
                      help="print file dependencies")
    parser.add_option(
        "--inorder", "-i", action="store_true", dest="in_order",
        help="processing in read order (default, can be omitted)")

    # verbosity options
    parser.add_option("-q", action="store_const", dest="verbosity", const=0,
                      help="quiet operation, suppressing warnings")
    parser.add_option("-v", action="count", dest="verbosity",
                      help="increase verbosity")
    parser.add_option("--verbosity", metavar='level', dest="verbosity",
                      type='int',
                      help=textwrap.dedent("""\
                      set verbosity level
                      0: quiet, suppressing warnings
                      1: default, showing warnings and error locations
                      2: show stack trace
                      3: log property definitions and usage on top level
                      4: log property definitions and usage on all levels"""))

    mappings = {}
    filtered_args = argv

    parser.set_defaults(just_deps=False, just_includes=False, verbosity=1)
    (options, pos_args) = parser.parse_args(filtered_args)
    if options.in_order:
        message(
            "xacro: in-order processing became default in ROS Melodic. "
            "You can drop the option.")
    options.in_order = True

    if len(pos_args) != 1:
        if require_input:
            parser.error("expected exactly one input file as argument")
        else:
            pos_args = [None]

    options.mappings = mappings
    return options, pos_args[0]



import threading
import time


class Ticker(threading.Thread):  # pragma nocover

    def __init__(self, period):
        super().__init__()

        self.sem = threading.Semaphore(0)
        self.done = False
        self.period = period
        print('in init', period)

    def wait(self):
        self.sem.acquire()

    def run(self):
        print('in run', self.period)
        start = time.time()
        stop = start
        while not self.done:
            stop += self.period
            # print('sleeping for', stop-start)
            start = time.time()
            time.sleep(stop - start)
            self.sem.release()

    def stop(self):
        self.done = True
        self.join()


if __name__ == "__main__":  # pragma nocover
    t = Ticker(0.1)

    tprev = time.time()
    tmax = 0
    tsum = 0
    N = 100

    t.start()
    for i in range(100):
        t.wait()
        tnow = time.time()
        dt = tnow - tprev
        tprev = tnow
        terr = abs(dt - 0.1) * 1000  # error in ms
        if i > 0:
            tsum += terr
            tmax = max(tmax, terr)
        print('.', end='', flush=True)
    t.stop()
    print(f"\nmean = {tsum/(N-1):.2f}, max = {tmax:.2f}")



#!/usr/bin/env python
"""
@author Micah Huth
"""

import importlib
import threading
import glob
import os
import platform
import warnings
from time import perf_counter, sleep
import imageio
from roboticstoolbox.backends.Connector import Connector
# from roboticstoolbox.robot.DHLink import DHLink
# from roboticstoolbox.robot.Robot import Robot as r
from roboticstoolbox import DHLink, DHRobot

GraphicsCanvas3D = None
GraphicsCanvas2D = None
GraphicalRobot = None
close_localhost_session = None


def _imports():  # pragma nocover
    global GraphicsCanvas3D
    global GraphicsCanvas2D
    global GraphicalRobot
    global close_localhost_session

    try:
        canvas = importlib.import_module(
            'roboticstoolbox.backends.VPython.canvas')
        GraphicsCanvas3D = canvas.GraphicsCanvas3D
        GraphicsCanvas2D = canvas.GraphicsCanvas2D

        graphicalrobot = importlib.import_module(
            'roboticstoolbox.backends.VPython.graphicalrobot')
        GraphicalRobot = graphicalrobot.GraphicalRobot

        common_functions = importlib.import_module(
            'roboticstoolbox.backends.VPython.common_functions')
        close_localhost_session = common_functions.close_localhost_session

    except ImportError:
        print(
            '\nYou must install the VPython component of the toolbox, do: \n'
            'pip install roboticstoolbox[vpython]\n\n')


class VPython(Connector):  # pragma nocover
    """
    Graphical backend using VPython

    VPython is a Python API that connects to a JavaScript/WebGL 3D graphics
    engine in a browser tab.  It supports many 3D graphical primitives
    including meshes, boxes, ellipsoids and lines. It can not render in
    full color.

    Example:

    .. code-block:: python
        :linenos:

        import roboticstoolbox as rtb

        robot = rtb.models.DH.Panda()  # create a robot

        pyplot = rtb.backends.VPython() # create a VPython backend
        pyplot.add(robot)              # add the robot to the backend
        robot.q = robot.qz             # set the robot configuration
        pyplot.step()                  # update the backend and graphical view

    :references:

        - https://vpython.org

    """
    # TODO be able to add ellipsoids (vellipse, fellipse)
    # TODO be able add lines (for end-effector paths)

    def __init__(self):
        """
        Open a localhost session with no canvases

        """
        super(VPython, self).__init__()

        _imports()

        # Init vars
        self.canvases = []
        # 2D array of [is_3d, height, width, title, caption, grid] per canvas
        self.canvas_settings = []
        self.robots = []
        self._recording = False
        self._recording_thread = None
        self._recording_fps = None
        self._thread_lock = threading.Lock()

        self._create_empty_session()

    def launch(
            self, is_3d=True, height=500, width=888,
            title='', caption='', grid=True,
            g_col=None):
        """
        Launch a graphical backend in a browser tab

        ``env = launch(args)` create a 3D scene in a new browser tab as
        defined by args, and returns a reference to the backend.

        """

        super().launch()

        self.canvas_settings.append(
            [is_3d, height, width, title, caption, grid, g_col])

        # Create the canvas with the given information
        if is_3d:
            self.canvases.append(
                GraphicsCanvas3D(height, width, title, caption,
                                 grid, g_col))
        else:
            self.canvases.append(
                GraphicsCanvas2D(height, width, title, caption,
                                 grid, g_col))

    def step(self, id, q=None, fig_num=0):
        """
        Update the graphical scene

        :param id: The Identification of the robot to remove. Can be either the
            DHRobot or GraphicalRobot
        :type id: :class:`~roboticstoolbox.robot.DHRobot.DHRobot`,
            :class:`roboticstoolbox.backends.VPython.graphics_robot.GraphicalRobot`
        :param q: The joint angles/configuration of the robot (Optional, if not
            supplied will use the stored q values).
        :type q: float ndarray(n)
        :param fig_num: The canvas index to delete the robot from, defaults to
            the initial one
        :type fig_num: int, optional
        :raises ValueError: Figure number must be between 0 and total number of
            canvases
        :raises TypeError: Input must be a DHLink or GraphicalRobot

        ``env.step(args)`` triggers an update of the 3D scene in the browser
        window referenced by ``env``.

        .. note::

            - Each robot in the scene is updated based on
              their control type (position, velocity, acceleration, or torque).
            - Upon acting, the other three of the four control types will be
              updated in the internal state of the robot object.
            - The control type is defined by the robot object, and not all
              robot objects support all control types.
            - Execution is blocked for the specified interval

        """

        super().step()

        if fig_num < 0 or fig_num >= len(self.canvases):
            raise ValueError(
                "Figure number must be between 0 and total number of canvases")

        # If DHRobot given
        if isinstance(id, DHRobot):
            robot = None
            # Find first occurrence of it that is in the correct canvas
            for i in range(len(self.robots)):
                if self.robots[i].robot is id and \
                        self.canvases[fig_num].is_robot_in_canvas(
                                                        self.robots[i]):
                    robot = self.robots[i]
                    break
            if robot is None:
                print("No robot")
                return
            else:
                poses = robot.fkine(q)
                robot.set_joint_poses(poses)
        # ElseIf GraphicalRobot given
        elif isinstance(id, GraphicalRobot):
            if self.canvases[fig_num].is_robot_in(id):
                poses = id.fkine(q)
                id.set_joint_poses(poses)
        # Else
        else:
            raise TypeError(
                "Input must be a Robot (or subclass) or "
                "GraphicalRobot, given {0}".format(type(id)))

    def reset(self):
        """
        Reset the graphical scene

        ``env.reset()`` triggers a reset of the 3D scene in the browser window
        referenced by ``env``. It is restored to the original state defined by
        ``launch()``.

        """

        super().reset()

        if len(self.canvases) > 0:
            # Clear localhost
            self.canvases[0].scene.append_to_caption('''
                <script type="text/javascript">
                    let gs = document.getElementById('glowscript');
                    gs.innerHTML = '';
                </script>
                ''')

            # Delete all sessions
            self.canvases = []

            self._create_empty_session()
            for settings in self.canvas_settings:
                # Create the canvas with the given information
                if settings[0]:
                    self.canvases.append(GraphicsCanvas3D(
                        settings[1], settings[2], settings[3],
                        settings[4], settings[5]))
                else:
                    self.canvases.append(GraphicsCanvas2D(
                        settings[1], settings[2], settings[3],
                        settings[4], settings[5]))

    def restart(self):
        """
        Restart the graphics display

        ``env.restart()`` triggers a restart of the browser view referenced by
        ``env``. It is closed and relaunched to the original state defined by
        ``launch()``.

        """

        super().restart()

        self.reset()

    def close(self):
        """
        Close the graphics display

        ``env.close()`` gracefully closes the browser tab browser view
        referenced by ``env``.

        """

        super().close()

        # Close session
        if len(self.canvases) > 0:
            # if a canvas made
            close_localhost_session(self.canvases[0])
        else:
            # No canvas, so make one
            temp = GraphicsCanvas2D()
            close_localhost_session(temp)

        self.canvases = []

    def add(self, fig_num, name, dhrobot):
        """
        Add a robot to the graphical scene

        :param fig_num: The canvas number to place the robot in
        :type fig_num: int
        :param name: The name of the robot
        :type name: `str`
        :param dhrobot: The ``DHRobot`` object (if applicable)
        :type dhrobot: class:`~roboticstoolbox.robot.DHRobot.DHRobot`, None
        :raises ValueError: Figure number must be between 0 and number of
            figures created
        :return: object id within visualizer
        :rtype: int

        ``id = env.add(robot)`` adds the ``robot`` to the graphical
            environment.

        .. note::

            - ``robot`` must be of an appropriate class.
            - Adds the robot object to a list of robots which will be updated
              when the ``step()`` method is called.

        """

        # TODO - name can come from the robot object, maybe an override name?
        #  Micah: "Name is used from robot class, unless robot is not given"

        # TODO - why dhrobot "if applicable"?
        #  Micah: "It's possible to create a graphical robot
        #  in VPython not using a robot class."

        # TODO - what about other classes of robot?
        #  Micah: "I use specific parameters in dhrobots.
        #  If they exist in other robot classes, it should work."

        # TODO - what about adding ellipsoids?

        super().add()

        # Sanity check input
        if fig_num < 0 or fig_num > len(self.canvases) - 1:
            raise ValueError(
                "Figure number must be between 0 and number "
                "of figures created")

        # Add robot to canvas
        self.robots.append(
            GraphicalRobot(self.canvases[fig_num], name, dhrobot))
        # self.canvases[fig_num].add_robot(self.robots[len(self.robots)-1])

    def remove(self, id, fig_num=0):
        """
        Remove a robot to the graphical scene

        :param id: The id of the robot to remove. Can be either the DHLink or
            GraphicalRobot
        :type id: class:`~roboticstoolbox.robot.DHRobot.DHRobot`,
                  class:`roboticstoolbox.backends.VPython.graphics_robot.GraphicalRobot`
        :param fig_num: The canvas index to delete the robot from, defaults to
             the initial one
        :type fig_num: int, optional
        :raises ValueError: Figure number must be between 0 and total number
            of canvases
        :raises TypeError: Input must be a DHLink or GraphicalRobot

        ``env.remove(robot)`` removes the ``robot`` from the graphical
            environment.

        """

        super().remove()

        if fig_num < 0 or fig_num >= len(self.canvases):
            raise ValueError(
                "Figure number must be between 0 and total number of canvases")

        # If DHLink given
        if isinstance(id, DHLink):
            robot = None
            # Find first occurrence of it that is in the correct canvas
            for i in range(len(self.robots)):
                if self.robots[i].seriallink.equal(id) and \
                        self.canvases[fig_num].is_robot_in(self.robots[i]):
                    robot = self.robots[i]
                    break
            if robot is None:
                return
            else:
                self.canvases[fig_num].delete_robot(robot)
        # ElseIf GraphicalRobot given
        elif isinstance(id, GraphicalRobot):
            if self.canvases[fig_num].is_robot_in(id):
                self.canvases[fig_num].delete_robot(id)
        # Else
        else:
            raise TypeError("Input must be a DHLink or GraphicalRobot")

    def hold(self):           # pragma: no cover
        '''
        hold() keeps the tab open i.e. stops the tab from closing once
        the main script has finished.

        '''

        while True:
            pass

    #
    # Public non-standard methods
    #
    def record_start(self, fps, scene_num=0):
        """
        Start recording screencaps of a scene
        """
        self._thread_lock.acquire()

        if not self._recording:
            print("VPython Recording...")
            if fps > 10:
                warnings.warn("The chosen recording fps ({0}) could result in lagging video quality."
                              "Consider lowering fps and robot speed (e.g. 5fps)".format(fps), RuntimeWarning)
            self._recording = True
            self._recording_fps = fps
            # Spawn a thread
            self._recording_thread = threading.Thread(target=self._record_scene, args=(scene_num, fps,))
            self._recording_thread.start()

        self._thread_lock.release()

    def record_stop(self, filename, save_fps=None):
        """
        Stop recording screencaps of a scene and combine them into a movie
        Save_fps is different to record fps. Will save the media file at the given save fps.
        """
        #
        self._thread_lock.acquire()
        if self._recording:
            self._recording = False
            print("VPython Recording Stopped...")
            print("VPython Recording Saving... DO NOT EXIT")
        else:
            self._thread_lock.release()
            return
        self._thread_lock.release()

        # Wait for thread to finish
        self._recording_thread.join()

        sleep(3)  # Quick sleep to ensure all downloads are done
        # (higher framerates can lag behind)

        # Get downloads directory
        opsys = platform.system()
        if opsys == 'Windows':  # Windows
            path_in = os.path.join(os.getenv('USERPROFILE'), 'downloads')

        elif opsys == 'Linux' or opsys == 'Darwin':  # Linux / Mac
            path_in = os.path.join(os.getenv('HOME'), 'downloads')

        else:  # Undefined OS
            # lets assume 'HOME' for now
            path_in = os.path.join(os.getenv('HOME'), 'downloads')

        fp_out = filename
        fp_in = path_in + "/vpython_*.png"

        files = [file for file in glob.glob(fp_in)]

        if save_fps is None:
            save_fps = self._recording_fps
        writer = imageio.get_writer(fp_out, fps=save_fps)

        for f in files:
            writer.append_data(imageio.imread(f))  # Add it to the video
            os.remove(f)  # Clean up file

        writer.close()

        print("VPython Recording Saved... It is safe to exit")

    #
    #  Private Methods
    #

    @staticmethod
    def _create_empty_session():
        """
        Create a canvas to ensure the localhost session has been opened.
        Then clear the browser tab
        """
        # Create a canvas to initiate the connection
        temp = GraphicsCanvas3D()

        # Delete the canvas to leave a blank screen
        temp.scene.append_to_caption('''
            <script type="text/javascript">
                let gs = document.getElementById('glowscript');
                gs.innerHTML = '';
            </script>
        ''')

    def _record_scene(self, scene_num, fps):
        """
        Thread-called function to continuously record screenshots
        """
        frame_num = 0
        if fps <= 0:
            raise ValueError("fps must be greater than 0.")
        f = 1 / fps

        self._thread_lock.acquire()
        recording = self._recording
        self._thread_lock.release()

        while recording:
            # Get current time
            t_start = perf_counter()

            # Take screenshot
            filename = "vpython_{:04d}.png".format(frame_num)
            self.canvases[scene_num].take_screenshot(filename)
            frame_num += 1

            # Get current time
            t_stop = perf_counter()

            # Wait for time of frame to finish
            # If saving takes longer than frame frequency, this while is skipped
            while t_stop - t_start < f:
                t_stop = perf_counter()

            self._thread_lock.acquire()
            recording = self._recording
            self._thread_lock.release()
